from . import world
from .constants import Constants, const
from .core import (Arg, From, FromArg, Get, Provide, Scope, Wiring, inject, wire)
from .factory import Factory, factory
from .implementation import implementation
from .lazy import LazyCall, LazyMethodCall
from .service import Service, service, ABCService
from .utils import is_compiled


try:
    from ._internal.scm_version import version as __version__
except ImportError:  # pragma: no cover
    __version__ = ''


__all__ = ['__version__', 'world', 'Get', 'From', 'FromArg', 'Provide',
           'const', 'Constants', 'inject', 'Arg', 'wire', 'Wiring', 'factory',
           'Factory', 'implementation', 'LazyCall', 'LazyMethodCall',
           'service', 'Service', 'ABCService', 'is_compiled']



import textwrap

from antidote import (Constants, Factory, From, LazyCall, LazyMethodCall, Provide,
                      Service, const, implementation, inject, world)
from antidote._compatibility.typing import Annotated
from antidote._internal.utils import short_id


class DebugTestCase:
    def __init__(self, value, expected: str, depth: int = -1, legend: bool = True):
        if expected.startswith('\n'):
            expected = textwrap.dedent(expected[1:])
        lines = expected.splitlines(keepends=True)
        while not lines[-1].strip():
            lines.pop()

        if legend:
            from antidote._internal.utils.debug import _LEGEND
            lines.append(_LEGEND)
        self.value = value
        self.expected = ''.join(lines)
        self.depth = depth


def assert_valid(*test_cases: DebugTestCase):
    for test_case in test_cases:
        assert world.debug(test_case.value,
                           depth=test_case.depth) == test_case.expected


def test_no_debug():
    def g():
        pass

    assert_valid(
        DebugTestCase(
            value=g,
            expected=f"""{g!r} is neither a dependency nor is anything injected.""",
            legend=False
        )
    )


def test_injections():
    prefix = "tests.world.test_debug.test_injections.<locals>"

    class MyService(Service):
        pass

    @inject(auto_provide=True)
    def f(s: MyService):
        return s

    @inject(auto_provide=True)
    async def g(s: MyService):
        return s

    assert_valid(
        DebugTestCase(
            value=f,
            expected=f"""
                {prefix}.f
                └── {prefix}.MyService
                    """
        ),
        DebugTestCase(
            value=g,
            expected=f"""
                {prefix}.g
                └── {prefix}.MyService
                    """
        )
    )


def test_implementation_debug():
    class Interface:
        pass

    prefix = "tests.world.test_debug.test_implementation_debug.<locals>"

    with world.test.new():
        class Dummy(Interface, Service):
            pass

        @implementation(Interface)
        def f():
            return Dummy

        assert_valid(DebugTestCase(
            value=Interface @ f,
            expected=f"""
                Permanent implementation: {prefix}.Interface @ {prefix}.f
                └── {prefix}.Dummy
                    """
        ))

        world.get(Interface @ f)
        assert_valid(DebugTestCase(
            value=Interface @ f,
            expected=f"""
                Permanent implementation: {prefix}.Interface @ {prefix}.f
                └── {prefix}.Dummy
                    """
        ))

    with world.test.new():
        class Dummy2(Interface, Service):
            pass

        undefined_expectations = f"""
            <∅> Implementation: {prefix}.Interface @ {prefix}.g
            └── {prefix}.Dummy2
        """

        @implementation(Interface, permanent=False)
        def g():
            return Dummy2

        assert_valid(DebugTestCase(value=Interface @ g,
                                   expected=undefined_expectations))

        world.get(Interface @ g)
        assert_valid(DebugTestCase(value=Interface @ g,
                                   expected=undefined_expectations))


def test_lazy_call_debug():
    prefix = "tests.world.test_debug.test_lazy_call_debug.<locals>"

    with world.test.new():
        def f():
            pass

        l1 = LazyCall(f)("arg", hello="world")
        assert_valid(
            DebugTestCase(
                value=l1,
                expected=f"""
                    Lazy: {prefix}.f(*('arg',), **{{'hello': 'world'}})  #{short_id(l1)}
                    """
            ),
            DebugTestCase(
                value=LazyCall(f, singleton=False)("arg", bye="world"),
                expected=f"""
                    <∅> Lazy: {prefix}.f(*('arg',), **{{'bye': 'world'}})
                    """
            ))

    with world.test.new():
        class MyService(Service):
            pass

        @inject
        def f(service: Provide[MyService]):
            pass

        l2 = LazyCall(f)
        assert_valid(
            DebugTestCase(
                value=l2,
                expected=f"""
                    Lazy: {prefix}.f()  #{short_id(l2)}
                    └── {prefix}.f
                        └── {prefix}.MyService
                    """
            ),
            DebugTestCase(
                value=LazyCall(f, singleton=False),
                expected=f"""
                    <∅> Lazy: {prefix}.f()
                    └── {prefix}.f
                        └── {prefix}.MyService
                    """
            )
        )


def test_unknown_debug():
    prefix = "tests.world.test_debug.test_unknown_debug.<locals>"

    with world.test.new():
        class Service1:
            pass

        class Dummy(Service):
            def __init__(self, s: Provide[Service1]):
                pass

        assert_valid(DebugTestCase(
            value=Dummy,
            expected=f"""
                {prefix}.Dummy
                └── /!\\ Unknown: {prefix}.Service1
            """
        ))


def test_singleton_debug():
    with world.test.new():
        world.test.singleton("test", 1)

        assert_valid(DebugTestCase(
            value="test",
            expected="""
                Singleton: 'test' -> 1
            """
        ))


def test_wiring_debug():
    prefix = "tests.world.test_debug.test_wiring_debug.<locals>"

    with world.test.new():
        class Service1(Service):
            pass

        # Not wired, but @inject
        class DummyA(Service):
            __antidote__ = Service.Conf(wiring=None)

            @inject
            def __init__(self, service: Provide[Service1]):
                pass

        assert_valid(DebugTestCase(
            value=DummyA,
            expected=f"""
                {prefix}.DummyA
                └── {prefix}.Service1
            """
        ))

        # Multiple injections
        class DummyB(Service):
            __antidote__ = Service.Conf()

            def __init__(self, s: Provide[Service1]):
                pass

            def get(self, s: Provide[Service1]):
                pass

        assert_valid(DebugTestCase(
            value=DummyB,
            expected=f"""
                {prefix}.DummyB
                ├── {prefix}.Service1
                └── Method: get
                    └── {prefix}.Service1
            """
        ))

        # Methods specified
        class DummyC(Service):
            __antidote__ = Service.Conf().with_wiring(methods=['__init__', 'get'])

            def __init__(self):
                pass

            def get(self, my_service: Provide[Service1]):
                pass

        assert_valid(
            DebugTestCase(
                value=DummyC,
                expected=f"""
                {prefix}.DummyC
                └── Method: get
                    └── {prefix}.Service1
                    """
            ))

        # No injections
        class DummyD(Service):
            def __init__(self):
                pass

            def get(self):
                pass

        assert_valid(
            DebugTestCase(
                value=DummyD,
                expected=f"""
                    {prefix}.DummyD
                    """
            ))


def test_multiline_debug():
    prefix = "tests.world.test_debug.test_multiline_debug.<locals>"

    with world.test.new():
        class MultilineService(Service):
            @classmethod
            def __antidote_debug_repr__(cls):
                return "Multiline\nService"

        class Dummy(Service):
            def __init__(self, s: Provide[MultilineService]):
                pass

        assert_valid(DebugTestCase(
            value=Dummy,
            expected=f"""
                {prefix}.Dummy
                └── Multiline
                    Service
            """
        ))


def test_lazy_method_debug():
    prefix = "tests.world.test_debug.test_lazy_method_debug.<locals>"

    with world.test.new():
        class MyService(Service):
            pass

        class Conf(Service):
            def fetch(self, value, service: Provide[MyService]):
                return value

            DATA = LazyMethodCall(fetch)
            KW = LazyMethodCall(fetch)(value='1')
            DATA2 = LazyMethodCall(fetch, singleton=False)
            KW2 = LazyMethodCall(fetch, singleton=False)(value='2')

        assert_valid(
            DebugTestCase(
                value=Conf.DATA,
                expected=f"""
                    Lazy Method: fetch()  #{short_id(Conf.__dict__['DATA'])}
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf
                    """
            ),
            DebugTestCase(
                value=Conf.KW,
                expected=f"""
        Lazy Method: fetch(*(), **{{'value': '1'}})  #{short_id(Conf.__dict__['KW'])}
        ├── {prefix}.Conf.fetch
        │   └── {prefix}.MyService
        └── {prefix}.Conf
                    """
            ),
            DebugTestCase(
                value=Conf.DATA2,
                expected=f"""
                    <∅> Lazy Method: fetch()
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf
                    """
            ),
            DebugTestCase(
                value=Conf.KW2,
                expected=f"""
                    <∅> Lazy Method: fetch(*(), **{{'value': '2'}})
                    ├── {prefix}.Conf.fetch
                    │   └── {prefix}.MyService
                    └── {prefix}.Conf
                    """
            )
        )


def test_constants_debug():
    prefix = "tests.world.test_debug.test_constants_debug.<locals>"

    with world.test.new():
        class Conf(Constants):
            TEST = const('1')

            def provide_const(self, key, arg):
                return key

        assert_valid(
            DebugTestCase(
                value=Conf.TEST,
                expected=f"""
                    {prefix}.Conf.TEST
                    """
            ))

    with world.test.new():
        class MyService(Service):
            pass

        class Conf(Constants):
            TEST = const('1')

            def provide_const(self, key, arg, service: Provide[MyService] = None):
                assert isinstance(service, MyService)
                return key

        assert_valid(
            DebugTestCase(
                value=Conf.TEST,
                expected=f"""
                    {prefix}.Conf.TEST
                    └── {prefix}.Conf.provide_const
                        └── {prefix}.MyService
                    """
            ))


def test_custom_scope():
    prefix = "tests.world.test_debug.test_custom_scope.<locals>"

    with world.test.new():
        dummy_scope = world.scopes.new(name='dummy')

        class MyService(Service):
            __antidote__ = Service.Conf(scope=dummy_scope)

        class BigService(Service):
            def __init__(self, my_service: Provide[MyService]):
                pass

        assert_valid(
            DebugTestCase(
                value=MyService,
                expected=f"""
                    <dummy> {prefix}.MyService
                    """
            ),
            DebugTestCase(
                value=BigService,
                expected=f"""
                    {prefix}.BigService
                    └──<dummy> {prefix}.MyService
                    """
            )
        )


def test_complex_debug():
    with world.test.new():
        class Interface:
            pass

        class Service1(Service):
            __antidote__ = Service.Conf(parameters=['test'])

        class Service2:
            def __init__(self, service1: Provide[Service1]):
                self.service1 = service1

        class BuildS2(Factory):
            __antidote__ = Factory.Conf(parameters=['option'])

            def __call__(self, service1: Provide[Service1] = None) -> Service2:
                return Service2(service1)

        @implementation(Interface)
        def impl():
            return Service4

        class Service3(Service):
            __antidote__ = Service.Conf(singleton=False).with_wiring(
                dependencies=dict(
                    i=Interface @ impl,
                    service2=Service2 @ BuildS2))

            def __init__(self,
                         service1: Provide[Service1],
                         service2: Service2,
                         i: Interface):
                self.service1 = service1
                self.service2 = service2
                self.i = i

            def get(self, service1: Provide[Service1]):
                pass

            X = LazyMethodCall(get)

        class Service4(Interface, Service):
            def __init__(self,
                         service1: Provide[Service1],
                         service2: Annotated[Service2, From(BuildS2)],
                         service3: Provide[Service3]):
                self.service1 = service1
                self.service2 = service2
                self.service3 = service3

        @inject
        def f(s: Provide[Service4]):
            pass

        @inject(dependencies=[Service1.parameterized(test=1),
                              Service2 @ BuildS2.parameterized(option=2)])
        def f_with_options(a, b):
            pass

        @inject
        def g(s: Provide[Service3], s4: Provide[Service4]):
            pass

        prefix = "tests.world.test_debug.test_complex_debug.<locals>"
        assert_valid(
            DebugTestCase(
                value=g,
                depth=0,
                expected=f"""
                    {prefix}.g
                    ├── {prefix}.Service4
                    └──<∅> {prefix}.Service3
                        """
            ),
            DebugTestCase(
                value=g,
                depth=1,
                expected=f"""
                    {prefix}.g
                    ├── {prefix}.Service4
                    └──<∅> {prefix}.Service3
                        """
            ),
            DebugTestCase(
                value=g,
                depth=2,
                expected=f"""
                    {prefix}.g
                    ├── {prefix}.Service4
                    │   ├──<∅> {prefix}.Service3
                    │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │   └── {prefix}.Service1
                    └──<∅> {prefix}.Service3
                        ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                        ├── {prefix}.Service2 @ {prefix}.BuildS2
                        └── {prefix}.Service1
                        """
            ),
            DebugTestCase(
                value=g,
                depth=3,
                expected=f"""
                {prefix}.g
                ├── {prefix}.Service4
                │   ├──<∅> {prefix}.Service3
                │   │   ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                │   │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                │   │   └── {prefix}.Service1
                │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                │   │   ├── {prefix}.BuildS2.__call__
                │   │   │   └── {prefix}.Service1
                │   │   └── {prefix}.BuildS2
                │   └── {prefix}.Service1
                └──<∅> {prefix}.Service3
                    ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                    │   └── {prefix}.Service4
                    ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │   ├── {prefix}.BuildS2.__call__
                    │   │   └── {prefix}.Service1
                    │   └── {prefix}.BuildS2
                    └── {prefix}.Service1
                        """
            ),
            DebugTestCase(
                value=g,
                expected=f"""
                {prefix}.g
                ├── {prefix}.Service4
                │   ├──<∅> {prefix}.Service3
                │   │   ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                │   │   │   └── /!\\ Cyclic dependency: {prefix}.Service4
                │   │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                │   │   │   ├── {prefix}.BuildS2.__call__
                │   │   │   │   └── {prefix}.Service1
                │   │   │   └── {prefix}.BuildS2
                │   │   └── {prefix}.Service1
                │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                │   │   ├── {prefix}.BuildS2.__call__
                │   │   │   └── {prefix}.Service1
                │   │   └── {prefix}.BuildS2
                │   └── {prefix}.Service1
                └──<∅> {prefix}.Service3
                    ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                    │   └── {prefix}.Service4
                    │       ├── /!\\ Cyclic dependency: {prefix}.Service3
                    │       ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │       │   ├── {prefix}.BuildS2.__call__
                    │       │   │   └── {prefix}.Service1
                    │       │   └── {prefix}.BuildS2
                    │       └── {prefix}.Service1
                    ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │   ├── {prefix}.BuildS2.__call__
                    │   │   └── {prefix}.Service1
                    │   └── {prefix}.BuildS2
                    └── {prefix}.Service1

                        """
            ),
            DebugTestCase(
                value=f,
                expected=f"""
                {prefix}.f
                └── {prefix}.Service4
                    ├──<∅> {prefix}.Service3
                    │   ├── Permanent implementation: {prefix}.Interface @ {prefix}.impl
                    │   │   └── /!\\ Cyclic dependency: {prefix}.Service4
                    │   ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │   │   ├── {prefix}.BuildS2.__call__
                    │   │   │   └── {prefix}.Service1
                    │   │   └── {prefix}.BuildS2
                    │   └── {prefix}.Service1
                    ├── {prefix}.Service2 @ {prefix}.BuildS2
                    │   ├── {prefix}.BuildS2.__call__
                    │   │   └── {prefix}.Service1
                    │   └── {prefix}.BuildS2
                    └── {prefix}.Service1
                        """
            ),
            DebugTestCase(
                value=f_with_options,
                expected=f"""
                {prefix}.f_with_options
                ├── {prefix}.Service2 @ {prefix}.BuildS2 with parameters={{'option': 2}}
                │   ├── {prefix}.BuildS2.__call__
                │   │   └── {prefix}.Service1
                │   └── {prefix}.BuildS2
                └── {prefix}.Service1 with parameters={{'test': 1}}
                    """
            )
        )



import sys

# Annotation
if sys.version_info < (3, 7):
    from typing import Callable, get_type_hints as _get_type_hints, cast
    from typing_extensions import Annotated, AnnotatedMeta

    # All of this is brittle, only working in our limited use case.
    def get_type_hints(obj: Callable[..., object], *, include_extras=False) -> object:
        # annotations cannot be removed in Python 3.6
        return _get_type_hints(obj)

    def get_origin(tp: object) -> object:
        if isinstance(tp, AnnotatedMeta):
            return Annotated
        return getattr(tp, "__origin__", None)

    def get_args(tp: object) -> tuple:
        return cast(tuple, getattr(tp, "__args__", tuple()))

elif sys.version_info < (3, 9):
    # get_origin and get_args must be retrieved from typing_extensions even for Python
    # 3.8 as the typing 3.8 implementation is not annotation-aware
    from typing_extensions import Annotated, get_type_hints, get_origin, get_args
else:
    from typing import Annotated, get_type_hints, get_origin, get_args

# Final / Protocol
if sys.version_info < (3, 8):
    from typing_extensions import final, Protocol
else:
    from typing import final, Protocol

# GenericMeta
if sys.version_info < (3, 7):
    from typing import GenericMeta
else:
    class GenericMeta(type):
        pass

__all__ = ['final', 'Protocol', 'GenericMeta', 'Annotated', 'get_type_hints',
           'get_origin', 'get_args']



from typing import Hashable, Optional

from antidote.core import (Container, DependencyValue, Provider)


class DummyIntProvider(Provider[int]):
    def __init__(self):
        super().__init__()
        self.original = None

    def exists(self, dependency: Hashable) -> bool:
        return isinstance(dependency, int)

    def provide(self, dependency: int, container: Container
                ) -> Optional[DependencyValue]:
        return DependencyValue(dependency * 2)

    def clone(self, keep_singletons_cache: bool):
        p = DummyIntProvider()
        p.original = self
        return p



from .factory import FactoryProvider
from .indirect import IndirectProvider
from .lazy import Lazy, LazyProvider
from .service import ServiceProvider
from .world_test import WorldTestProvider

__all__ = ['FactoryProvider', 'IndirectProvider', 'Lazy', 'LazyProvider',
           'ServiceProvider', 'WorldTestProvider']



from contextlib import contextmanager

import pytest

from antidote import Scope, world
from antidote.core.exceptions import DuplicateDependencyError


class Service:
    pass


@contextmanager
def empty_context():
    yield


@pytest.mark.parametrize('context', [empty_context(), world.test.clone()])
def test_invalid_root_world(context):
    with context:
        with pytest.raises(RuntimeError, match=".*test world.*"):
            world.test.singleton(Service, '1')

        with pytest.raises(RuntimeError, match=".*test world.*"):
            @world.test.factory()
            def build() -> Service:
                return Service()


def test_singleton():
    with world.test.empty():
        world.test.singleton("singleton", 12342)
        assert world.get("singleton") == 12342

        world.test.singleton({
            "singleton2": 89,
            "singleton3": 123
        })
        assert world.get("singleton2") == 89
        assert world.get("singleton3") == 123


def test_invalid_singleton():
    with world.test.empty():
        with pytest.raises(TypeError):
            world.test.singleton(1)

        with pytest.raises(TypeError):
            world.test.singleton(dict(), 1)


def test_duplicate_singletons():
    with world.test.empty():
        world.test.singleton("singleton", 12342)

        with pytest.raises(DuplicateDependencyError, match=".*singleton.*12342.*"):
            world.test.singleton("singleton", 1)

        with pytest.raises(DuplicateDependencyError, match=".*singleton.*12342.*"):
            world.test.singleton({"singleton": 1})


def test_factory():
    with world.test.empty():
        s = Service()

        @world.test.factory(Service)
        def build():
            return s

        assert world.get(Service) is s


def test_factory_from_annotation():
    with world.test.empty():
        s = Service()

        @world.test.factory()
        def build() -> Service:
            return s

        assert world.get(Service) is s


def test_factory_singleton():
    with world.test.empty():
        @world.test.factory(Service)
        def build():
            return Service()

        assert world.get(Service) is world.get(Service)

    with world.test.empty():
        @world.test.factory(Service, singleton=True)
        def build2():
            return Service()

        assert world.get(Service) is world.get(Service)

    with world.test.empty():
        @world.test.factory(Service, scope=Scope.singleton())
        def build3():
            return Service()

        assert world.get(Service) is world.get(Service)


def test_factory_no_scope():
    with world.test.empty():
        @world.test.factory(Service, singleton=False)
        def build():
            return Service()

        assert world.get(Service) is not world.get(Service)


def test_factory_scope():
    with world.test.empty():
        scope = world.scopes.new(name='dummy')

        @world.test.factory(Service, scope=scope)
        def build():
            return Service()

        s = world.get(Service)
        assert world.get(Service) is s

        world.scopes.reset(scope)
        assert world.get(Service) is not s


def test_invalid_factory():
    with world.test.empty():
        with pytest.raises(TypeError, match=".*factory.*callable.*"):
            world.test.factory()(object())

        with pytest.raises(ValueError, match="(?i).*either.*dependency.*type hint.*"):
            @world.test.factory()
            def build():
                return Service()



import functools
import inspect
from typing import Callable, Hashable

from ._internal import API
from ._providers.indirect import ImplementationDependency


@API.private
class ImplementationWrapper:
    def __init__(self,
                 wrapped: Callable[..., object],
                 implementation_dependency: ImplementationDependency) -> None:
        self.__wrapped__ = wrapped
        self.__implementation_dependency = implementation_dependency
        functools.wraps(wrapped, updated=())(self)

    def __call__(self, *args: object, **kwargs: object) -> object:
        return self.__wrapped__(*args, **kwargs)

    def __rmatmul__(self, left_operand: Hashable) -> object:
        if left_operand is not self.__implementation_dependency.interface:
            interface = self.__implementation_dependency.interface
            raise ValueError(f"Unsupported interface {left_operand}, "
                             f"expected {interface}")
        return self.__implementation_dependency

    def __getattr__(self, item: str) -> object:
        return getattr(self.__wrapped__, item)


@API.private
def validate_provided_class(dependency: Hashable, *, expected: type) -> None:
    from ._providers.factory import FactoryDependency
    from ._providers.service import Parameterized
    from ._providers.indirect import ImplementationDependency

    cls: object = dependency
    if isinstance(cls, Parameterized):
        cls = cls.wrapped
    if isinstance(cls, FactoryDependency):
        cls = cls.output
    if isinstance(cls, ImplementationDependency):
        cls = cls.interface

    if not (isinstance(cls, type) and inspect.isclass(cls)):
        raise TypeError(f"{dependency} does not provide any class")

    if not issubclass(cls, expected):
        raise TypeError(f"Expected a subclass of {expected}, not {cls}")



from typing import Optional

from ._internal import API
from .core import Scope
from .core.injection import validate_injection

__all__ = ['is_compiled', 'validate_injection', 'validated_scope']


@API.public
def is_compiled() -> bool:
    """
    Whether current Antidote implementations is the compiled (Cython) version or not
    """
    from ._internal.wrapper import compiled
    return compiled


@API.public
def validated_scope(scope: Optional[Scope] = Scope.sentinel(),
                    singleton: Optional[bool] = None,
                    *,
                    default: Optional[Scope]) -> Optional[Scope]:
    """
    Validates given arguments and ensures consistency between singleton and scope.

    Top-level APIs exposes both singleton and scope arguments. Users can choose either,
    but not both. This function does all validation necessary in those APIs.

    .. doctest:: utils_validated_scope

        >>> from antidote import Scope, world
        >>> from antidote.utils import validated_scope
        >>> custom_scope = world.scopes.new(name='custom')
        >>> validated_scope(singleton=None, default=custom_scope)
        Scope(name='custom')
        >>> validated_scope(singleton=True, default=None)
        Scope(name='singleton')
        >>> validated_scope(scope=Scope.singleton(), default=None)
        Scope(name='singleton')

    Args:
        scope: Scope argument to validate. Neutral value is :py:meth:`.Scope.sentinel`.
        singleton: Singleton argument to validate. Neutral value is :py:obj:`None`.
        default: Default value to use if both scope and singleton have neutral values.

    Returns:
        Scope defined by either :code:`singleton` or :code:`scope`.

    """
    if not (scope is None or isinstance(scope, Scope)):
        raise TypeError(f"scope must be a Scope or None, not {type(scope)}")
    if not (default is None or isinstance(default, Scope)):
        raise TypeError(f"default must be a Scope or None, not {type(default)}")

    if isinstance(singleton, bool):
        if scope is not Scope.sentinel():
            raise TypeError("Use either singleton or scope argument, not both.")
        return Scope.singleton() if singleton else None
    if singleton is not None:
        raise TypeError(f"singleton must be a boolean or None, "
                        f"not {type(singleton)}")

    return default if scope is Scope.sentinel() else scope



import collections.abc as c_abc
import inspect
from typing import (Any, Callable, Dict, Hashable, Iterable, List, Mapping, TYPE_CHECKING,
                    Union, cast, Set)

from .exceptions import DoubleInjectionError
from .._internal import API
from .._internal.argspec import Arguments
from .._internal.wrapper import (Injection, InjectionBlueprint, build_wrapper,
                                 get_wrapped, is_wrapper)

if TYPE_CHECKING:
    from .injection import DEPENDENCIES_TYPE, AUTO_PROVIDE_TYPE

AnyF = Union[Callable[..., Any], staticmethod, classmethod]


@API.private
def raw_inject(f: AnyF,
               dependencies: 'DEPENDENCIES_TYPE',
               auto_provide: 'AUTO_PROVIDE_TYPE',
               strict_validation: bool) -> AnyF:
    if not isinstance(strict_validation, bool):
        raise TypeError(f"strict_validation must be a boolean, "
                        f"not {type(strict_validation)}")
    if inspect.isclass(f):
        # User-friendlier error for classes.
        raise TypeError("Classes cannot be wrapped with @inject. "
                        "Consider using @wire")

    real_f = f.__func__ if isinstance(f, (classmethod, staticmethod)) else f
    if is_wrapper(f):
        raise DoubleInjectionError(get_wrapped(f))
    if is_wrapper(real_f):
        raise DoubleInjectionError(get_wrapped(real_f))
    if not inspect.isfunction(real_f):
        raise TypeError(f"wrapped object {f} is neither a function "
                        f"nor a (class/static) method")

    blueprint = _build_injection_blueprint(
        arguments=Arguments.from_callable(f),
        dependencies=dependencies,
        auto_provide=auto_provide,
        strict_validation=strict_validation
    )
    # If nothing can be injected, just return the existing function without
    # any overhead.
    if blueprint.is_empty():
        return f

    wrapped_real_f = build_wrapper(blueprint=blueprint, wrapped=real_f)
    if isinstance(f, staticmethod):
        return staticmethod(wrapped_real_f)
    if isinstance(f, classmethod):
        return classmethod(wrapped_real_f)
    return cast(AnyF, wrapped_real_f)


@API.private
def _build_injection_blueprint(arguments: Arguments,
                               dependencies: 'DEPENDENCIES_TYPE',
                               auto_provide: 'AUTO_PROVIDE_TYPE',
                               strict_validation: bool
                               ) -> InjectionBlueprint:
    """
    Construct a InjectionBlueprint with all the necessary information about
    the arguments for dependency injection. Storing it avoids significant
    execution overhead./

    Used by inject()
    """
    annotated = _build_from_annotations(arguments)
    explicit_dependencies = _build_from_dependencies(arguments, dependencies,
                                                     strict_validation)
    auto_provided = _build_auto_provide(arguments, auto_provide, annotated,
                                        strict_validation)

    resolved_dependencies: List[object] = [
        annotated.get(arg.name,
                      explicit_dependencies.get(arg.name,
                                                auto_provided.get(arg.name)))
        for arg in arguments
    ]

    return InjectionBlueprint(tuple(
        Injection(arg_name=arg.name,
                  required=not arg.has_default,
                  dependency=dependency)
        for arg, dependency in zip(arguments, resolved_dependencies)
    ))


@API.private
def _build_from_annotations(arguments: Arguments) -> Dict[str, Hashable]:
    from ._annotations import extract_annotated_arg_dependency
    arg_to_dependency = {}
    for arg in arguments:
        dependency = extract_annotated_arg_dependency(arg)
        if dependency is not None:
            arg_to_dependency[arg.name] = dependency

    return arg_to_dependency


@API.private
def _build_from_dependencies(arguments: Arguments,
                             dependencies: 'DEPENDENCIES_TYPE',
                             strict_validation: bool
                             ) -> Dict[str, Hashable]:
    from .injection import Arg
    if dependencies is None:
        arg_to_dependency: Mapping[str, Hashable] = {}
    elif callable(dependencies):
        arg_to_dependency = {arg.name: dependencies(Arg(arg.name,
                                                        arg.type_hint,
                                                        arg.type_hint_with_extras))
                             for arg in arguments.without_self}
    elif isinstance(dependencies, c_abc.Mapping):
        _check_valid_arg_names("dependencies",
                               dependencies.keys(),
                               arguments,
                               strict_validation)
        arg_to_dependency = dependencies
    elif isinstance(dependencies, c_abc.Iterable) and not isinstance(dependencies, str):
        # convert to Tuple in case we cannot iterate more than once.
        dependencies = tuple(dependencies)
        if strict_validation and len(arguments.without_self) < len(dependencies):
            raise ValueError(f"More dependencies ({dependencies}) were provided "
                             f"than arguments ({arguments})")
        arg_to_dependency = {arg.name: dependency
                             for arg, dependency
                             in zip(arguments.without_self, dependencies)}
    else:
        raise TypeError(f'Only a mapping or a iterable is supported for '
                        f'dependencies, not {type(dependencies)!r}')

    # Remove any None as they would hide type_hints.
    return {k: v for k, v in arg_to_dependency.items() if v is not None}


@API.private
def _build_auto_provide(arguments: Arguments,
                        auto_provide: 'AUTO_PROVIDE_TYPE',
                        annotated: Dict[str, Hashable],
                        strict_validation: bool
                        ) -> Dict[str, Hashable]:
    from ._annotations import extract_auto_provided_arg_dependency

    if isinstance(auto_provide, bool):
        def is_auto_provided(__cls: type) -> bool:
            return cast(bool, auto_provide)
    elif isinstance(auto_provide, c_abc.Iterable):
        # convert to Tuple in case we cannot iterate more than once.
        auto_provide = set(auto_provide)
        for cls in auto_provide:
            if not (isinstance(cls, type) and inspect.isclass(cls)):
                raise TypeError(f"auto_provide must be a boolean or an iterable of "
                                f"classes, but contains {cls!r} which is not a class.")

        def is_auto_provided(__cls: type) -> bool:
            return __cls in cast(Set[type], auto_provide)
    elif callable(auto_provide):
        is_auto_provided = auto_provide
    else:
        raise TypeError(f"auto_provide must be a boolean, an iterable of classes, "
                        f"or a function not {type(auto_provide)}.")

    auto_provided: Dict[str, Union[type, Hashable]] = {}
    for arg in arguments.without_self:
        dependency = extract_auto_provided_arg_dependency(arg)
        if dependency is not None and is_auto_provided(dependency):
            auto_provided[arg.name] = dependency

    provided_dependencies = set(auto_provided.values()).union(annotated.values())
    if strict_validation \
            and isinstance(auto_provide, set) \
            and not auto_provide.issubset(provided_dependencies):
        raise ValueError(f"Some auto_provide dependencies ({auto_provide}) are not "
                         f"present in the function. Found: {provided_dependencies}\n"
                         f"Either ensure that auto_provide matches the function type "
                         f"hints or consider specifying strict_validation=False")

    return auto_provided


@API.private
def _check_valid_arg_names(param: str,
                           names: Iterable[str],
                           arguments: Arguments,
                           strict_validation: bool) -> None:
    for name in names:
        if not isinstance(name, str):
            raise TypeError(f"{param} expected an argument name (str), "
                            f"not {name!r} ({type(name)})")
        if arguments.has_self and name == arguments[0].name:
            raise ValueError(f"Cannot inject first argument in {param} "
                             f"({arguments[0]!r}) of a method / @classmethod.")

        if strict_validation and name not in arguments:
            raise ValueError(f"Unknown argument in {param}: {name!r}")



from antidote.core import DependencyDebug, DependencyValue, Scope


def test_dependency_value():
    ref = DependencyValue("test", scope=Scope.singleton())
    assert ref == DependencyValue("test", scope=Scope.singleton())
    assert ref != DependencyValue("test2", scope=Scope.singleton())
    assert ref != DependencyValue("test", scope=None)


def test_dependency_debug():
    ref = DependencyDebug("info",
                          scope=Scope.singleton(),
                          wired=[1],
                          dependencies=[2])
    assert ref == DependencyDebug("info",
                                  scope=Scope.singleton(),
                                  wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug("info2",
                                  scope=Scope.singleton(),
                                  wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug("info",
                                  scope=None,
                                  wired=[1],
                                  dependencies=[2])
    assert ref != DependencyDebug("info",
                                  scope=Scope.singleton(),
                                  wired=[10],
                                  dependencies=[2])
    assert ref != DependencyDebug("info",
                                  scope=Scope.singleton(),
                                  wired=[1],
                                  dependencies=[20])



from typing import Dict

from .. import API


@API.private
class SlotsRepr:
    __slots__ = ()

    def __repr__(self) -> str:
        slots_attrs = []
        for cls in type(self).__mro__:
            for name in getattr(cls, '__slots__', []):
                attr = f"_{cls.__name__}{name}" if name.startswith('__') else name
                slots_attrs.append(f'{name}={getattr(self, attr)!r}')
        return f"{type(self).__name__}({', '.join(slots_attrs)})"


@API.private
class SlotRecord(SlotsRepr):
    """
    Used in similar fashion to data classes. Used whenever mutability is still needed
    """
    __slots__ = ()

    def __init__(self, *args: object, **kwargs: object) -> None:
        super().__init__()
        attrs: Dict[str, object] = dict(zip(self.__slots__, args))
        attrs.update(kwargs)
        for attr, value in attrs.items():
            setattr(self, attr, value)



import os

from antidote import __version__, is_compiled


def test_is_compiled():
    assert isinstance(is_compiled(), bool)
    assert is_compiled() == (os.environ.get("ANTIDOTE_COMPILED") == "true")


def test_version():
    assert isinstance(__version__, str)



"""
Utilities used by world, mostly for syntactic sugar.
"""
from typing import (Generic, Hashable, TYPE_CHECKING, Type, TypeVar, Union,
                    overload)

from . import API
from .state import current_container
from .utils import Default
from .utils.immutable import Immutable, ImmutableGenericMeta
from .utils.meta import FinalMeta
from .._compatibility.typing import Protocol, final
from ..core._annotations import extract_annotated_dependency
from ..core.container import RawContainer
from ..core.exceptions import DependencyNotFoundError

T = TypeVar('T')

if TYPE_CHECKING:
    from .._constants import Const


@API.private
class SupportsRMatmul(Protocol):
    def __rmatmul__(self, type_hint: object) -> Hashable:
        pass  # pragma: no cover


@API.private
@final
class LazyDependency(Immutable, Generic[T], metaclass=ImmutableGenericMeta):
    """
    Recommended usage is to usage :py:func:`..world.lazy`:

    .. doctest:: core_utils_dependency

        >>> from antidote import Service, world
        ... class MyService(Service):
        ...     pass
        >>> port = world.lazy[MyService]()
        >>> port.get()
        <MyService ...>

    """
    __slots__ = ('unwrapped', '_type')
    unwrapped: Hashable
    """Actual dependency to be retrieved"""
    _type: Type[T]

    def __init__(self,
                 __dependency: Hashable,
                 expected_type: Type[T]) -> None:
        """
        Args:
            __dependency: actual dependency to be retrieved.
        """
        super().__init__(__dependency, expected_type)

    def get(self) -> T:
        """
        Returns:
            dependency value retrieved from :py:mod:`~..world`.
        """
        from antidote import world
        value = world.get(self.unwrapped)

        if not isinstance(value, self._type):
            raise TypeError(f"Dependency is not an instance of {self._type}, "
                            f"but {type(value)}")

        return value


@API.private
@final
class WorldGet(metaclass=FinalMeta):
    @overload
    def __call__(self,  # noqa: E704
                 __dependency: 'Const[T]',
                 *,
                 default: Union[T, Default] = Default.sentinel
                 ) -> T:
        ...  # pragma: no cover

    @overload
    def __call__(self,  # noqa: E704
                 __dependency: Hashable,
                 *,
                 default: object = Default.sentinel
                 ) -> object:
        ...  # pragma: no cover

    def __call__(self,
                 __dependency: Hashable,
                 *,
                 default: object = Default.sentinel
                 ) -> object:
        dependency = extract_annotated_dependency(__dependency)
        try:
            return current_container().get(dependency)
        except DependencyNotFoundError:
            if default is not Default.sentinel:
                return default
            raise

    def __getitem__(self, tpe: Type[T]) -> 'TypedWorldGet[T]':
        return TypedWorldGet(tpe)


@API.private
@final
class TypedWorldGet(Generic[T], Immutable, metaclass=ImmutableGenericMeta):
    __slots__ = ('__type',)
    __type: Type[T]

    def __call__(self,
                 __dependency: Hashable = None,
                 *,
                 default: Union[T, Default] = Default.sentinel) -> T:
        if not isinstance(default, (self.__type, Default)):
            raise TypeError(f"default is not an instance of {self.__type}, "
                            f"but {type(default)}")

        if __dependency is None:
            dependency: object = self.__type
        else:
            dependency = extract_annotated_dependency(__dependency)
        try:
            value = current_container().get(dependency)
        except DependencyNotFoundError:
            if default is not Default.sentinel:
                return default
            raise

        if not isinstance(value, self.__type):
            raise TypeError(f"Dependency is not an instance of {self.__type}, "
                            f"but {type(value)}")

        return value

    def __matmul__(self, other: SupportsRMatmul) -> T:
        return self.__call__(self.__type @ other)


@API.private
@final
class WorldLazy(metaclass=FinalMeta):
    def __call__(self, __dependency: Hashable) -> LazyDependency[object]:
        return LazyDependency(__dependency, object)

    def __getitem__(self, tpe: Type[T]) -> 'TypedWorldLazy[T]':
        return TypedWorldLazy(tpe)


@API.private
@final
class TypedWorldLazy(Generic[T], Immutable, metaclass=ImmutableGenericMeta):
    __slots__ = ('__type',)
    __type: Type[T]

    def __call__(self, __dependency: Hashable = None) -> LazyDependency[T]:
        return LazyDependency[T](self.__type if __dependency is None else __dependency,
                                 self.__type)

    def __matmul__(self, other: SupportsRMatmul) -> LazyDependency[T]:
        return self.__call__(self.__type @ other)


@API.private
def new_container() -> RawContainer:
    """ default new container in Antidote """

    from .._providers import (LazyProvider, ServiceProvider,
                              IndirectProvider, FactoryProvider)

    container = RawContainer()
    container.add_provider(FactoryProvider)
    container.add_provider(LazyProvider)
    container.add_provider(IndirectProvider)
    container.add_provider(ServiceProvider)

    return container



import functools
import inspect
from typing import Callable, Dict, Set, Tuple, Type, TypeVar, cast

from .container import RawProvider
from .exceptions import FrozenWorldError
from .._compatibility.typing import GenericMeta
from .._internal import API

_FREEZE_ATTR_NAME = "__antidote__freeze_sensitive"


# TODO: Inheriting GenericMeta for Python 3.6. To be removed ASAP.
@API.private
class ProviderMeta(GenericMeta):
    def __new__(mcs: 'Type[ProviderMeta]',
                name: str,
                bases: Tuple[type, ...],
                namespace: Dict[str, object],
                abstract: bool = False,
                **kwargs: object
                ) -> 'ProviderMeta':
        # Every method which does not the have the does_not_freeze decorator
        # is considered
        raw_methods = {"clone", "provide", "exists", "maybe_provide", "debug",
                       "maybe_debug"}
        attrs: Set[str] = {attr for attr in namespace.keys() if
                           not attr.startswith("__")}
        for attr in (attrs - raw_methods):
            method = namespace[attr]
            if inspect.isfunction(method) and callable(method):
                if getattr(method, _FREEZE_ATTR_NAME, True):
                    namespace[attr] = _make_wrapper(attr, method)

        cls = cast(
            ProviderMeta,
            super().__new__(mcs, name, bases, namespace, **kwargs)  # type: ignore
        )
        assert getattr(cls, "__antidote__") is None

        return cls


F = TypeVar('F', bound=Callable[..., object])


@API.private
def _make_wrapper(attr: str, method: F) -> F:
    @functools.wraps(method)
    def wrapped_method(self: RawProvider, *args: object, **kwargs: object) -> object:
        try:
            with self._bound_container_locked(freezing=True):
                # If you have a TypeError traceback pointing here you probably have
                # a mismatch between the arguments and the wrapped method signature.
                return method(self, *args, **kwargs)
        except FrozenWorldError:
            raise FrozenWorldError(
                f"Method {attr} could not be called in a frozen "
                f"world with args={args} and kwargs={kwargs}"
            )

    return cast(F, wrapped_method)



from typing import Callable, Optional, TypeVar, Union

import pytest

from antidote import From, FromArg, Get, Provide
from antidote._compatibility.typing import Annotated
from antidote._internal.argspec import Arguments
from antidote.core._annotations import (AntidoteAnnotation,
                                        extract_annotated_arg_dependency,
                                        extract_annotated_dependency,
                                        extract_auto_provided_arg_dependency)

T = TypeVar('T')


class Dummy:
    pass


class Maker:
    def __rmatmul__(self, other):
        assert other is Dummy
        return Maker


def test_invalid_from_arg():
    with pytest.raises(TypeError, match=".*function.*"):
        FromArg(object())


@pytest.mark.parametrize('type_hint,expected', [
    pytest.param(type_hint, expected,
                 id=str(type_hint).replace('typing.', '').replace(f"{__name__}.", ""))
    for type_hint, expected in [
        (Annotated[Dummy, object()], None),
        (Dummy, None),
        (str, None),
        (T, None),
        (Union[str, Dummy], None),
        (Union[str, Dummy, int], None),
        (Optional[Union[str, Dummy]], None),
        (Callable[..., Dummy], None),
        (Provide[Dummy], Dummy),
        (Annotated[Dummy, From(Maker())], Maker),
        (Annotated[Dummy, FromArg(lambda arg: arg.name * 2)], 'xx'),
        (Annotated[Dummy, Get('something')], 'something'),  # noqa: F821
    ]
])
def test_extract_explicit_arg_dependency(type_hint, expected):
    def f(x: type_hint):
        pass

    arguments = Arguments.from_callable(f)
    assert extract_annotated_arg_dependency(arguments[0]) == expected

    def g(x: type_hint = None):
        pass

    arguments = Arguments.from_callable(g)
    assert extract_annotated_arg_dependency(arguments[0]) == expected


@pytest.mark.parametrize('type_hint,expected', [
    pytest.param(type_hint, expected,
                 id=str(type_hint).replace('typing.', '').replace(f"{__name__}.", ""))
    for type_hint, expected in [
        (Dummy, Dummy),
        (Annotated[Dummy, object()], Dummy),
        (str, None),
        (T, None),
        (Union[str, Dummy], None),
        (Union[str, Dummy, int], None),
        (Optional[Union[str, Dummy]], None),
        (Callable[..., Dummy], None),
        (Provide[Dummy], None),
        (Annotated[Dummy, From(Maker())], None),
        (Annotated[Dummy, FromArg(lambda arg: arg.name * 2)], None),
        (Annotated[Dummy, Get('something')], None),  # noqa: F821
    ]
])
def test_extract_auto_provided_arg_dependency(type_hint, expected):
    def f(x: type_hint):
        pass

    arguments = Arguments.from_callable(f)
    assert extract_auto_provided_arg_dependency(arguments[0]) == expected

    def g(x: type_hint = None):
        pass

    arguments = Arguments.from_callable(g)
    assert extract_auto_provided_arg_dependency(arguments[0]) == expected


@pytest.mark.parametrize('type_hint,expected', [
    pytest.param(type_hint, expected,
                 id=str(type_hint).replace('typing.', '').replace(f"{__name__}.", ""))
    for type_hint, expected in [
        (Provide[Dummy], Dummy),
        (Annotated[Dummy, object()], Dummy),
        (str, str),
        (T, T),
        (Union[str, Dummy], Union[str, Dummy]),
        (Annotated[Dummy, From(Maker())], Maker),
        (Annotated[Dummy, Get('something')], 'something')  # noqa: F821
    ]
])
def test_extract_annotated_dependency(type_hint, expected):
    assert extract_annotated_dependency(type_hint) == expected


def test_multiple_antidote_annotations():
    type_hint = Annotated[Dummy, Get('dummy'), Get('dummy')]  # noqa: F821

    def f(x: type_hint):
        pass

    arguments = Arguments.from_callable(f)
    with pytest.raises(TypeError):
        extract_annotated_arg_dependency(arguments[0])

    with pytest.raises(TypeError):
        extract_annotated_dependency(type_hint)


def test_unknown_antidote_annotations():
    type_hint = Annotated[Dummy, AntidoteAnnotation()]

    def f(x: type_hint):
        pass

    arguments = Arguments.from_callable(f)
    with pytest.raises(TypeError):
        extract_annotated_arg_dependency(arguments[0])

    with pytest.raises(TypeError):
        extract_annotated_dependency(type_hint)


@pytest.mark.parametrize('type_hint', [
    pytest.param(type_hint,
                 id=str(type_hint).replace('typing.', '').replace(f"{__name__}.", ""))
    for type_hint in [
        Annotated[Dummy, FromArg(lambda arg: arg.name * 2)]
    ]
])
def test_argument_only_annotations(type_hint):
    with pytest.raises(TypeError):
        extract_annotated_dependency(type_hint)



import itertools
from typing import Optional, Sequence, Union

import pytest

from antidote import From, FromArg, Get, world
from antidote._compatibility.typing import Annotated
from antidote.core.annotations import Provide
from antidote.core.injection import inject, validate_injection
from antidote.exceptions import DependencyNotFoundError, DoubleInjectionError

SENTINEL = "SENTINEL"


class DoesNotRaise:
    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class MyService:
    pass


class AnotherService:
    pass


@pytest.mark.parametrize('kwargs, expectation', [
    pytest.param(kwargs, expectation, id=str(kwargs))
    for kwargs, expectation in [
        (dict(), DoesNotRaise()),
        (dict(dependencies=lambda arg: arg.name), DoesNotRaise()),
        (dict(dependencies=dict(x='x')), DoesNotRaise()),
        (dict(dependencies=['x']), DoesNotRaise()),
        (dict(auto_provide=True), DoesNotRaise()),
        (dict(dependencies="{arg_name}"),
         pytest.raises(TypeError, match=".*dependencies.*str.*")),
        (dict(dependencies=1),
         pytest.raises(TypeError, match=".*dependencies.*int.*")),
        (dict(auto_provide=1),
         pytest.raises(TypeError, match=".*auto_provide.*int.*")),
        (dict(auto_provide=['x']),
         pytest.raises(TypeError, match=".*auto_provide.*x.*")),
    ]
])
def test_validate_injection(kwargs, expectation):
    with expectation:
        @inject(**kwargs)
        def f(x: MyService):
            pass

    with expectation:
        validate_injection(**kwargs)


@pytest.fixture(params=[inject])
def injector(request):
    return request.param


def test_simple(injector):
    @injector
    def f(x: Provide[MyService]):
        return x

    with world.test.empty():
        s = MyService()
        world.test.singleton(MyService, s)
        assert s == f()


no_SENTINEL_injection = [
    pytest.param((None, None),
                 dict(),
                 id='nothing'),
    pytest.param((MyService, None),
                 dict(dependencies=dict(first=MyService)),
                 id='dependencies:dict-first'),
    pytest.param((MyService, None),
                 dict(dependencies=(MyService,)),
                 id='dependencies:tuple-first'),
    pytest.param((None, MyService),
                 dict(dependencies=dict(second=MyService)),
                 id='dependencies:dict-second'),
    pytest.param((None, MyService),
                 dict(dependencies=(None, MyService)),
                 id='dependencies:tuple-second'),
    pytest.param(('first', 'second'),
                 dict(dependencies=lambda arg: arg.name),
                 id='dependencies:callable'),
    pytest.param((MyService, MyService),
                 dict(dependencies=lambda arg: MyService),
                 id='dependencies:callable2'),
    pytest.param((None, None),
                 dict(dependencies=lambda s: None),
                 id='dependencies:callable3')
]


@pytest.mark.parametrize('expected,kwargs', no_SENTINEL_injection)
def test_without_type_hints(injector, expected, kwargs):
    @injector(**kwargs)
    def f(first=SENTINEL, second=SENTINEL):
        return first, second

    class A:
        @injector(**kwargs)
        def method(self, first=SENTINEL, second=SENTINEL):
            return first, second

        @injector(**kwargs)
        @classmethod
        def class_method(cls, first=SENTINEL, second=SENTINEL):
            return first, second

        @injector(**kwargs)
        @staticmethod
        def static_method(first=SENTINEL, second=SENTINEL):
            return first, second

        @classmethod
        @injector(**kwargs)
        def klass(cls, first=SENTINEL, second=SENTINEL):
            return first, second

    with world.test.empty():
        world.test.singleton({
            MyService: MyService(),
            AnotherService: AnotherService(),
            'first': object(),
            'second': object(),
            'prefix:first': object(),
            'prefix:second': object()
        })

        expected = tuple((
            world.get(d) if d is not None else SENTINEL
            for d in expected
        ))
        (first, second) = expected
        a, b = object(), object()

        assert (a, second) == f(first=a)
        assert (a, second) == A().method(first=a)
        assert (a, second) == A.class_method(first=a)
        assert (a, second) == A.static_method(first=a)
        assert (a, second) == A.klass(first=a)

        assert (first, b) == f(second=b)
        assert (first, b) == A().method(second=b)
        assert (first, b) == A.class_method(second=b)
        assert (first, b) == A.static_method(second=b)
        assert (first, b) == A.klass(second=b)

        assert (a, second) == f(a)
        assert (a, second) == A().method(a)
        assert (a, second) == A.class_method(a)
        assert (a, second) == A.static_method(a)
        assert (a, second) == A.klass(a)

        assert (a, b) == f(a, b)
        assert (a, b) == A().method(a, b)
        assert (a, b) == A.class_method(a, b)
        assert (a, b) == A.static_method(a, b)
        assert (a, b) == A.klass(a, b)


@pytest.mark.parametrize('expected,kwargs', no_SENTINEL_injection + [
    pytest.param((MyService, AnotherService),
                 dict(auto_provide=True),
                 id='auto_provide:True'),
    pytest.param((None, None),
                 dict(auto_provide=False),
                 id='auto_provide:False'),
    pytest.param((MyService, None),
                 dict(auto_provide=[MyService]),
                 id='auto_provide:list-first'),
    pytest.param((None, AnotherService),
                 dict(auto_provide=[AnotherService]),
                 id='auto_provide:list-second'),
    pytest.param((None, None),
                 dict(auto_provide=[]),
                 id='auto_provide:list-None'),
    pytest.param((MyService, AnotherService),
                 dict(auto_provide=lambda cls: True),
                 id='auto_provide:func-True'),
    pytest.param((None, None),
                 dict(auto_provide=lambda cls: False),
                 id='auto_provide:func-False'),
    pytest.param((MyService, None),
                 dict(auto_provide=lambda cls: issubclass(cls, MyService)),
                 id='auto_provide:func-first'),
    pytest.param((None, AnotherService),
                 dict(auto_provide=lambda cls: issubclass(cls, AnotherService)),
                 id='auto_provide:func-second'),
    pytest.param((MyService, MyService),
                 dict(dependencies=dict(second=MyService), auto_provide=True),
                 id='auto_provide&dependencies:second'),
    pytest.param((None, MyService),
                 dict(dependencies=dict(second=MyService), auto_provide=[AnotherService]),
                 id='auto_provide&dependencies:second')
])
def test_with_auto_provide(injector, expected, kwargs):
    @injector(**kwargs)
    def f(first: MyService = SENTINEL, second: AnotherService = SENTINEL):
        return first, second

    class A:
        @injector(**kwargs)
        def method(self, first: MyService = SENTINEL, second: AnotherService = SENTINEL):
            return first, second

        @injector(**kwargs)
        @classmethod
        def class_method(cls, first: MyService = SENTINEL,
                         second: AnotherService = SENTINEL):
            return first, second

        @injector(**kwargs)
        @staticmethod
        def static_method(first: MyService = SENTINEL, second: AnotherService = SENTINEL):
            return first, second

        @classmethod
        @injector(**kwargs)
        def klass(cls, first: MyService = SENTINEL, second: AnotherService = SENTINEL):
            return first, second

    with world.test.empty():
        world.test.singleton({
            MyService: MyService(),
            AnotherService: AnotherService(),
            'first': "first",
            'second': "second",
            'prefix:first': "prefix:first",
            'prefix:second': "prefix:second"
        })

        expected = tuple((
            world.get(d) if d is not None else SENTINEL
            for d in expected
        ))
        (first, second) = expected
        a, b = "a", "b"

        assert (a, second) == f(first=a)
        assert (a, second) == A().method(first=a)
        assert (a, second) == A.class_method(first=a)
        assert (a, second) == A.static_method(first=a)
        assert (a, second) == A.klass(first=a)

        assert (first, b) == f(second=b)
        assert (first, b) == A().method(second=b)
        assert (first, b) == A.class_method(second=b)
        assert (first, b) == A.static_method(second=b)
        assert (first, b) == A.klass(second=b)

        assert (a, second) == f(a)
        assert (a, second) == A().method(a)
        assert (a, second) == A.class_method(a)
        assert (a, second) == A.static_method(a)
        assert (a, second) == A.klass(a)

        assert (a, b) == f(a, b)
        assert (a, b) == A().method(a, b)
        assert (a, b) == A.class_method(a, b)
        assert (a, b) == A.static_method(a, b)
        assert (a, b) == A.klass(a, b)


@pytest.mark.parametrize(
    'expected, kwargs',
    [
        pytest.param((MyService, None),
                     dict(),
                     id='nothing'),
        pytest.param((MyService, None),
                     dict(dependencies=dict(first=AnotherService)),
                     id='dependencies:dict-first'),
        pytest.param((MyService, None),
                     dict(dependencies=(AnotherService,)),
                     id='dependencies:tuple-first'),
        pytest.param((MyService, MyService),
                     dict(dependencies=dict(second=MyService)),
                     id='dependencies:dict-second'),
        pytest.param((MyService, MyService),
                     dict(dependencies=(None, MyService)),
                     id='dependencies:tuple-second'),
        pytest.param((MyService, 'second'),
                     dict(dependencies=lambda arg: arg.name),
                     id='dependencies:callable'),
        pytest.param((MyService, AnotherService),
                     dict(dependencies=lambda arg: AnotherService),
                     id='dependencies:callable2'),
        pytest.param((MyService, None),
                     dict(dependencies=lambda arg: None),
                     id='dependencies:callable3'),
        pytest.param((MyService, AnotherService),
                     dict(auto_provide=True),
                     id='auto_provide:True'),
        pytest.param((MyService, None),
                     dict(auto_provide=False),
                     id='auto_provide:False'),
        pytest.param((MyService, None),
                     dict(auto_provide=[MyService]),
                     id='auto_provide:list-first'),
        pytest.param((MyService, AnotherService),
                     dict(auto_provide=[AnotherService]),
                     id='auto_provide:list-second'),
        pytest.param((MyService, None),
                     dict(auto_provide=lambda cls: issubclass(cls, MyService)),
                     id='auto_provide:func-first'),
        pytest.param((MyService, AnotherService),
                     dict(auto_provide=lambda cls: issubclass(cls, AnotherService)),
                     id='auto_provide:func-second'),
    ]
)
def test_with_provide(injector, expected, kwargs):
    @injector(**kwargs)
    def f(first: Provide[MyService] = SENTINEL, second: AnotherService = SENTINEL):
        return first, second

    class A:
        @injector(**kwargs)
        def method(self,
                   first: Provide[MyService] = SENTINEL,
                   second: AnotherService = SENTINEL):
            return first, second

        @injector(**kwargs)
        @classmethod
        def class_method(cls,
                         first: Provide[MyService] = SENTINEL,
                         second: AnotherService = SENTINEL):
            return first, second

        @injector(**kwargs)
        @staticmethod
        def static_method(first: Provide[MyService] = SENTINEL,
                          second: AnotherService = SENTINEL):
            return first, second

        @classmethod
        @injector(**kwargs)
        def klass(cls,
                  first: Provide[MyService] = SENTINEL,
                  second: AnotherService = SENTINEL):
            return first, second

    with world.test.empty():
        world.test.singleton({MyService: MyService(),
                              AnotherService: AnotherService(),
                              'first': object(),
                              'second': object(),
                              'prefix:first': object(),
                              'prefix:second': object()})

        expected = tuple((
            world.get(d) if d is not None else SENTINEL
            for d in expected
        ))
        (first, second) = expected
        a, b = object(), object()

        assert (a, second) == f(first=a)
        assert (a, second) == A().method(first=a)
        assert (a, second) == A.class_method(first=a)
        assert (a, second) == A.static_method(first=a)
        assert (a, second) == A.klass(first=a)

        assert (first, b) == f(second=b)
        assert (first, b) == A().method(second=b)
        assert (first, b) == A.class_method(second=b)
        assert (first, b) == A.static_method(second=b)
        assert (first, b) == A.klass(second=b)

        assert (a, second) == f(a)
        assert (a, second) == A().method(a)
        assert (a, second) == A.class_method(a)
        assert (a, second) == A.static_method(a)
        assert (a, second) == A.klass(a)

        assert (a, b) == f(a, b)
        assert (a, b) == A().method(a, b)
        assert (a, b) == A.class_method(a, b)
        assert (a, b) == A.static_method(a, b)
        assert (a, b) == A.klass(a, b)


@pytest.mark.parametrize('type_hint',
                         # builtins
                         [str, int, float, set, list, dict, complex, type, tuple, bytes,
                          bytearray]
                         # typing
                         + [Optional, Sequence]
                         # not a class / weird stuff
                         + [1, lambda x: x, object()])
def test_ignored_type_hints(injector, type_hint):
    @injector(auto_provide=True)
    def f(x: type_hint):
        pass

    with world.test.empty():
        world.test.singleton(type_hint, object())
        with pytest.raises(TypeError):
            f()


def test_none_optional_support(injector):
    class Dummy:
        pass

    @injector
    def f(x: Provide[MyService] = None):
        return x

    @injector
    def g(x: Optional[Provide[MyService]] = None):
        return x

    with world.test.empty():
        s = MyService()
        world.test.singleton(Dummy, Dummy())
        world.test.singleton(MyService, s)
        assert f() is s
        assert g() is s

    with world.test.empty():
        world.test.singleton(Dummy, Dummy())
        assert f() is None
        assert g() is None


@pytest.mark.parametrize('auto_provide', [True, False])
def test_none_optional_support_auto_provide(injector, auto_provide):
    class Dummy:
        pass

    @injector(auto_provide=auto_provide)
    def f(x: MyService = None):
        return x

    @injector(auto_provide=auto_provide)
    def g(x: Optional[MyService] = None):
        return x

    @injector(auto_provide=auto_provide)
    def h(x: Union[MyService, Dummy]):
        pass

    with world.test.empty():
        s = MyService()
        world.test.singleton(Dummy, Dummy())
        world.test.singleton(MyService, s)
        expected = s if auto_provide else None
        assert f() is expected
        assert g() is expected
        with pytest.raises(TypeError):
            h()

    with world.test.empty():
        world.test.singleton(Dummy, Dummy())
        assert f() is None
        assert g() is None


def test_annotations(injector):
    class Dummy:
        pass

    with world.test.empty():
        @injector
        def custom_annotated(x: Annotated[Dummy, object()]):
            return x

        world.test.singleton(Dummy, Dummy())
        with pytest.raises(TypeError):
            custom_annotated()

    with world.test.empty():
        @injector
        def optional_annotated(x: Annotated[Dummy, object()] = None):
            return x

        world.test.singleton(Dummy, Dummy())
        assert optional_annotated() is None

    with world.test.empty():
        @injector
        def get(x: Annotated[Dummy, Get('dummy')]):  # noqa: F821
            return x

        world.test.singleton('dummy', Dummy())
        assert get() is world.get('dummy')

    with world.test.empty():
        @injector
        def optional_get(
                x: Optional[Annotated[Dummy, Get('dummy')]] = None):  # noqa: F821, E501
            return x

        world.test.singleton('dummy', Dummy())
        assert optional_get() is world.get('dummy')

    with world.test.empty():
        @injector
        def from_arg(x: Annotated[Dummy, FromArg(lambda arg: arg.name)]):
            return x

        world.test.singleton('x', Dummy())
        assert from_arg() is world.get('x')

    with world.test.empty():
        class Maker:
            def __rmatmul__(self, other):
                return 'dummy'

        @injector
        def get(x: Annotated[Dummy, From(Maker())]):
            return x

        world.test.singleton('dummy', Dummy())
        assert get() is world.get('dummy')


def test_multiple_antidote_annotations(injector):
    class Dummy:
        pass

    class Maker:
        def __rmatmul__(self, other):
            return 'dummy'

    annotations = [
        Get('dummy'),
        From(Maker()),
        FromArg(lambda arg: arg.name),
    ]
    for (a, b) in itertools.combinations(annotations, 2):
        with pytest.raises(TypeError):
            @injector
            def custom_annotated(x: Annotated[Dummy, a, b]):
                return x


@pytest.fixture(params=['function',
                        'method',
                        'classmethod',
                        'classmethod-after',
                        'staticmethod'])
def injected_method_with(request, injector):
    kind = request.param

    def builder(**kwargs):
        if kind == 'function':
            @injector(**kwargs)
            def f(x: MyService):
                return x

            return f
        else:
            class Dummy:
                @injector(**kwargs)
                def method(self, x: MyService):
                    return x

                @classmethod
                @injector(**kwargs)
                def class_method(cls, x: MyService):
                    return x

                @injector(**kwargs)
                @classmethod
                def class_method_after(cls, x: MyService):
                    return x

                @injector(**kwargs)
                @staticmethod
                def static_method(x: MyService):
                    return x

            if kind == 'method':
                return Dummy().method
            elif kind == 'classmethod':
                return Dummy.class_method
            elif kind == 'classmethod-after':
                return Dummy.class_method_after
            elif kind == 'staticmethod':
                return Dummy.static_method
            else:
                raise RuntimeError()

    return builder


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(TypeError, match=".*dependencies.*"),
                     dict(dependencies=object()),
                     id="dependencies:unsupported-type"),
        pytest.param(pytest.raises(TypeError),
                     dict(dependencies={1: 'x'}),
                     id="dependencies:invalid-key-type"),
        pytest.param(pytest.raises(TypeError, match=".*auto_provide.*"),
                     dict(auto_provide=object()),
                     id="auto_provide:unsupported-type"),
        pytest.param(pytest.raises(TypeError, match=".*strict_validation.*"),
                     dict(strict_validation=object()),
                     id="auto_provide:unsupported-type")
    ]
)
def test_invalid_args(injected_method_with, expectation, kwargs):
    with expectation:
        injected_method_with(**kwargs)()

    with expectation:
        validate_injection(**kwargs)


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(TypeError),
                     dict(),
                     id="nothing"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=(MyService,)),
                     id="dependencies:tuple"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=dict(x=MyService)),
                     id="dependencies:dict"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=lambda arg: MyService),
                     id="dependencies:callable"),
        pytest.param(pytest.raises(TypeError),
                     dict(auto_provide=False),
                     id="auto_provide:False"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(auto_provide=True),
                     id="auto_provide:True"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(auto_provide=[MyService]),
                     id="auto_provide:list")
    ]
)
def test_unknown_dependency(injected_method_with, expectation, kwargs):
    with expectation:
        injected_method_with(**kwargs)()


def test_unknown_provide(injector):
    @injector
    def f(x: Provide[MyService]):
        return x

    with pytest.raises(DependencyNotFoundError):
        f()


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(ValueError, match=".*y.*"),
                     dict(auto_provide=[AnotherService]),
                     id="auto_provide"),
        pytest.param(pytest.raises(ValueError, match=".*y.*"),
                     dict(auto_provide=[MyService, AnotherService]),
                     id="auto_provide2"),
        pytest.param(pytest.raises(ValueError),
                     dict(dependencies=(None, None)),
                     id="dependencies:too-much-arguments"),
        pytest.param(pytest.raises(ValueError, match=".*unknown.*"),
                     dict(dependencies=dict(unknown=DependencyNotFoundError)),
                     id="dependencies:unknown-dict-key"),
    ]
)
@pytest.mark.parametrize('strict', [True, False])
def test_strict_validation(injected_method_with, injector, expectation, kwargs, strict):
    kwargs['strict_validation'] = strict
    if strict:
        with expectation:
            injected_method_with(**kwargs)
    else:
        injected_method_with(**kwargs)(object())


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(ValueError, match=".*self.*"),
                     dict(dependencies=dict(self='x')),
                     id="dependencies"),
    ]
)
def test_cannot_inject_self(injector, expectation, kwargs):
    with expectation:
        class A:
            @injector(**kwargs)
            def method(self, x=None):
                return x

        A()

    with expectation:
        class A:
            @injector(**kwargs)
            @classmethod
            def classmethod(self, x=None):
                return x

        A()

    with expectation:
        class A:
            @classmethod
            @injector(**kwargs)
            def classmethod(self, x=None):
                return x

        A()


def test_no_injections():
    def f(x):
        return x

    # When nothing can be injected, the same function should be returned
    assert inject(f) is f


def test_double_injection(injector):
    # When the function has already its arguments injected, the same function should
    # be returned
    with pytest.raises(DoubleInjectionError,
                       match=".*<locals>.f.*"):
        @injector
        @injector(auto_provide=True)
        def f(x: MyService):
            return x

    with pytest.raises(DoubleInjectionError,
                       match=".*StaticmethodA.*"):
        class StaticmethodA:
            @injector
            @staticmethod
            @injector(auto_provide=True)
            def f(x, y: MyService):
                pass

    with pytest.raises(DoubleInjectionError,
                       match=".*StaticmethodB.*"):
        class StaticmethodB:
            @injector
            @injector(auto_provide=True)
            @staticmethod
            def f(x, y: MyService):
                pass

    with pytest.raises(DoubleInjectionError,
                       match=".*ClassmethodA.*"):
        class ClassmethodA:
            @injector
            @classmethod
            @injector(auto_provide=True)
            def f(cls, x: MyService):
                pass

    with pytest.raises(DoubleInjectionError,
                       match=".*ClassmethodB.*"):
        class ClassmethodB:
            @injector
            @injector(auto_provide=True)
            @classmethod
            def f(cls, x: MyService):
                pass

    with pytest.raises(DoubleInjectionError,
                       match=".*Method.*"):
        class Method:
            @injector
            @injector(auto_provide=True)
            def f(self, x: MyService):
                pass


def test_invalid_inject(injector):
    with pytest.raises(TypeError):
        @injector
        class Dummy:
            pass

    with pytest.raises(TypeError):
        injector(1)


def test_static_class_method(injector):
    class Dummy:
        @injector(auto_provide=True)
        @staticmethod
        def static(x: MyService):
            pass

        @injector(auto_provide=True)
        @classmethod
        def klass(cls, x: MyService):
            pass

    assert isinstance(Dummy.__dict__['static'], staticmethod)
    assert isinstance(Dummy.__dict__['klass'], classmethod)


@pytest.mark.asyncio
async def test_async(injector):
    with world.test.empty():
        service = MyService()
        another_service = AnotherService()
        world.test.singleton({
            MyService: service,
            AnotherService: another_service
        })

        @injector
        async def f(x: Provide[MyService]):
            return x

        res = await f()
        assert res is service

        class Dummy:
            @injector
            async def method(self, x: Provide[MyService]):
                return x

            @injector
            @classmethod
            async def klass(cls, x: Provide[MyService]):
                return x

            @injector
            @staticmethod
            async def static(x: Provide[MyService]):
                return x

        dummy = Dummy()
        print(dir(dummy))
        res = await dummy.method()
        assert res is service
        res = await dummy.klass()
        assert res is service
        res = await dummy.static()
        assert res is service
        res = await Dummy.klass()
        assert res is service
        res = await Dummy.static()
        assert res is service

        @injector(dependencies=(MyService, AnotherService))
        async def f(x, y):
            return x, y

        res = await f()
        assert res == (service, another_service)
        res = await f(y=1)
        assert res == (service, 1)
        res = await f(1)
        assert res == (1, another_service)
        res = await f(1, 2)
        assert res == (1, 2)

        @injector(dependencies=(MyService, 'unknown'))
        async def f(x, y):
            return x, y

        with pytest.raises(DependencyNotFoundError, match='.*unknown.*'):
            await f()

        @injector(dependencies=(MyService,))
        async def f(x, y):
            return x, y

        with pytest.raises(TypeError):
            await f()

        @injector(dependencies=(MyService, 'unknown'))
        async def f(x, y=None):
            return x, y

        res = await f()
        assert res == (service, None)


def test_dependencies_shortcut():
    with world.test.empty():
        world.test.singleton(MyService, MyService())

        @inject([MyService])
        def f(x):
            return x

        assert f() is world.get[MyService]()

        @inject(dict(x=MyService))
        def g(x):
            return x

        assert g() is world.get[MyService]()

        with pytest.raises(TypeError):
            @inject("test")
            def h(x):
                return x

        with pytest.raises(TypeError):
            @inject([MyService], dependencies=dict())
            def h2(x):
                return x

        with pytest.raises(TypeError):
            @inject(dict(x=MyService), dependencies=dict())
            def h3(x):
                return x



"""
Test only that the wrapper behaves nicely in all cases.
Injection itself is tested through inject.
"""
import inspect
from typing import Any, List, Tuple

import pytest

from antidote import world
from antidote._internal.wrapper import Injection, InjectionBlueprint, build_wrapper
from antidote.exceptions import DependencyNotFoundError

A = object()
B = object()
C = object()


@pytest.fixture(autouse=True, scope='module')
def empty_world():
    with world.test.empty():
        world.test.singleton('x', A)
        yield


def easy_wrap(func=None,
              arg_dependency: List[Tuple[str, bool, Any]] = tuple()):
    def wrapper(func):
        return build_wrapper(
            blueprint=InjectionBlueprint(tuple([
                Injection(arg_name, required, dependency)
                for arg_name, required, dependency in arg_dependency
            ])),
            wrapped=func
        )

    return func and wrapper(func) or wrapper


arg_self_x = [('self', True, None), ('x', True, 'x')]
arg_cls_x = [('cls', True, None), ('x', True, 'x')]
arg_x = [('x', True, 'x')]


class Dummy:
    @easy_wrap(arg_dependency=arg_self_x)
    def method(self, x):
        return self, x

    @classmethod
    @easy_wrap(arg_dependency=arg_cls_x)
    def class_after(cls, x):
        return cls, x

    @staticmethod
    @easy_wrap(arg_dependency=arg_x)
    def static_after(x):
        return x


class Dummy2:
    def method(self, x):
        return self, x


Dummy2.method = easy_wrap(Dummy2.__dict__['method'], arg_self_x)


@easy_wrap(arg_dependency=arg_x)
def f(x):
    return x


d = Dummy()
d2 = Dummy2()


@pytest.mark.parametrize(
    'expected, func',
    [
        pytest.param(A, f,
                     id='func'),

        pytest.param((B, A), Dummy.method,
                     id='method'),
        pytest.param((Dummy, A), Dummy.class_after,
                     id='classmethod after'),
        pytest.param(A, Dummy.static_after,
                     id='staticmethod after'),

        pytest.param((d, A), d.method,
                     id='instance method'),
        pytest.param((Dummy, A), d.class_after,
                     id='instance classmethod after'),
        pytest.param(A, d.static_after,
                     id='instance staticmethod after'),

        pytest.param((d2, A), d2.method,
                     id='post:instance method'),
        pytest.param((B, A), Dummy2.method,
                     id='post:method')
    ]
)
def test_wrapper(expected, func):
    if expected == (B, A):
        assert expected == func(B, A)
        assert expected == func(B)
        assert (B, C) == func(B, C)
        assert (B, C) == func(B, x=C)
    else:
        assert expected == func(A)
        assert expected == func()

        if isinstance(expected, tuple):
            new_expected = (expected[0], C)
        else:
            new_expected = C

        assert new_expected == func(C)
        assert new_expected == func(x=C)


def test_classmethod_wrapping():
    def class_method(cls):
        pass

    class A:
        method = easy_wrap(classmethod(class_method))

    assert class_method == A.__dict__['method'].__func__
    assert A == A.method.__self__


def test_required_dependency_not_found():
    @easy_wrap(arg_dependency=[('x', True, 'unknown')])
    def f(x):
        return x

    with pytest.raises(DependencyNotFoundError):
        f()


def test_dependency_not_found():
    @easy_wrap(arg_dependency=[('x', False, 'unknown')])
    def f(x):
        return x

    with pytest.raises(TypeError):
        f()


def test_multiple_injections():
    xx = object()
    yy = object()
    zz = object()

    @easy_wrap(arg_dependency=[('x', True, 'xx'),
                               ('y', True, 'yy'),
                               ('z', False, 'zz')])
    def f(x, y, z=zz):
        return x, y, z

    world.test.singleton(dict(xx=xx, yy=yy))
    assert (xx, yy, zz) == f()
    assert (xx, A, zz) == f(y=A)
    assert (xx, yy, A) == f(z=A)
    assert (A, yy, zz) == f(x=A)
    assert (A, yy, B) == f(A, z=B)

    with pytest.raises(TypeError):
        f(A, x=A)


def g():
    pass


def g_with_annotations(x: int):
    pass


def g_with_defaults(x, y=3):
    pass


def g_with_kw_only(x, *, y):
    pass


class G:
    def method(self):
        pass

    def method_with_annotations(self, x: int):
        pass

    def method_with_defaults(self, x, y=3):
        pass

    def method_with_kw_only(self, x, *, y):
        pass

    @staticmethod
    def cls():
        pass

    @staticmethod
    def cls_with_annotations(cls, x: int):
        pass

    @staticmethod
    def cls_with_defaults(cls, x, y=3):
        pass

    @staticmethod
    def cls_with_kw_only(cls, x, *, y):
        pass

    @staticmethod
    def static():
        pass

    @staticmethod
    def static_with_annotations(x: int):
        pass

    @staticmethod
    def static_with_defaults(x, y=3):
        pass

    @staticmethod
    def static_with_kw_only(x, *, y):
        pass


@pytest.mark.parametrize(
    'original,wrapped',
    [
        pytest.param(g,
                     easy_wrap(g),
                     id='function'),
        pytest.param(g_with_annotations,
                     easy_wrap(g_with_annotations,
                               arg_dependency=[('x', True, None)]),
                     id='function:__annotations__'),
        pytest.param(g_with_defaults,
                     easy_wrap(g_with_defaults,
                               arg_dependency=[('x', True, None), ('y', False, None)]),
                     id='function:__defaults__'),
        pytest.param(g_with_kw_only,
                     easy_wrap(g_with_kw_only,
                               arg_dependency=[('x', True, None), ('y', True, None)]),
                     id='function:__kwdefaults__'),
        pytest.param(G.method,
                     easy_wrap(G.method),
                     id='method'),
        pytest.param(G.method_with_annotations,
                     easy_wrap(G.method_with_annotations,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None)]),
                     id='method:__annotations__'),
        pytest.param(G.method_with_defaults,
                     easy_wrap(G.method_with_defaults,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None),
                                               ('y', False, None)]),
                     id='method:__defaults__'),
        pytest.param(G.method_with_kw_only,
                     easy_wrap(G.method_with_kw_only,
                               arg_dependency=[('self', True, None),
                                               ('x', True, None),
                                               ('y', True, None)]),
                     id='method:__kwdefaults__'),
        pytest.param(G.cls,
                     easy_wrap(G.cls),
                     id='classmethod'),
        pytest.param(G.cls_with_annotations,
                     easy_wrap(G.cls_with_annotations,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None)]),
                     id='classmethod:__annotations__'),
        pytest.param(G.cls_with_defaults,
                     easy_wrap(G.cls_with_defaults,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None),
                                               ('y', False, None)]),
                     id='classmethod:__defaults__'),
        pytest.param(G.cls_with_kw_only,
                     easy_wrap(G.cls_with_kw_only,
                               arg_dependency=[('cls', True, None),
                                               ('x', True, None),
                                               ('y', True, None)]),
                     id='classmethod:__kwdefaults__'),
        pytest.param(G.static,
                     easy_wrap(G.static),
                     id='static'),
        pytest.param(G.static_with_annotations,
                     easy_wrap(G.static_with_annotations,
                               arg_dependency=[('x', True, None)]),
                     id='static:__annotations__'),
        pytest.param(G.static_with_defaults,
                     easy_wrap(G.static_with_defaults,
                               arg_dependency=[('x', True, None), ('y', False, None)]),
                     id='static:__defaults__'),
        pytest.param(G.static_with_kw_only,
                     easy_wrap(G.static_with_kw_only,
                               arg_dependency=[('x', True, None), ('y', True, None)]),
                     id='static:__kwdefaults__')
    ]
)
def test_wrapped_attributes(original, wrapped):
    assert original is wrapped.__wrapped__
    assert original.__module__ == wrapped.__module__
    assert original.__name__ == wrapped.__name__
    assert original.__qualname__ == wrapped.__qualname__
    assert original.__doc__ == wrapped.__doc__
    assert original.__annotations__ == wrapped.__annotations__
    assert original.__defaults__ == wrapped.__defaults__
    assert original.__kwdefaults__ == wrapped.__kwdefaults__

    try:
        func = original.__func__
    except AttributeError:
        with pytest.raises(AttributeError):
            wrapped.__func__
    else:
        assert func is wrapped.__func__

    try:
        func = original.__self__
    except AttributeError:
        with pytest.raises(AttributeError):
            wrapped.__self__
    else:
        assert func is wrapped.__self__

    assert inspect.signature(original) == inspect.signature(wrapped)


def f():
    pass


async def async_f():
    pass


@pytest.mark.parametrize('func', [f, async_f])
def test_custom_attributes(func):
    func.attr = "test"
    func.another_attr = "another_attr"
    wrapped = easy_wrap(func)
    assert "test" == wrapped.attr

    wrapped.attr2 = "test2"
    assert "test2" == wrapped.attr2
    # After setting a new attribute, original ones should still be accessible.
    assert "test" == wrapped.attr

    wrapped.attr = "overridden test"
    assert "overridden test" == wrapped.attr

    # You should be able to remove new attributes
    del wrapped.attr2

    with pytest.raises(AttributeError):
        wrapped.attr2

    # but not existing ones
    with pytest.raises(AttributeError):
        del wrapped.another_attr


@pytest.mark.asyncio
async def test_async_wrapper():
    with world.test.empty():
        world.test.singleton(dict(a=A, b=B, c=C))

        @easy_wrap(arg_dependency=[('x', True, 'a')])
        async def f(x):
            return x

        res = await f()
        assert res == A

        class Dummy:
            @easy_wrap(arg_dependency=[('self', True, None), ('x', True, 'a')])
            async def method(self, x):
                return x

            @classmethod
            @easy_wrap(arg_dependency=[('cls', True, None), ('x', True, 'a')])
            async def klass(cls, x):
                return x

            @staticmethod
            @easy_wrap(arg_dependency=[('x', True, 'a')])
            async def static(x):
                return x

        d = Dummy()
        print(d.__dict__)
        res = await d.method()
        assert res == A
        res = await d.klass()
        assert res == A
        res = await d.static()
        assert res == A
        res = await Dummy.klass()
        assert res == A
        res = await Dummy.static()
        assert res == A

        @easy_wrap(arg_dependency=[('x', True, 'a'),
                                   ('y', True, 'b'),
                                   ('z', False, 'unknown')])
        async def f(x, y, z=None):
            return x, y, z

        (x, y, z) = await f()
        assert x == A
        assert y == B
        assert z is None

        @easy_wrap(arg_dependency=[('x', True, 'a'),
                                   ('y', True, 'b'),
                                   ('z', True, 'unknown')])
        async def f(x, y, z):
            pass

        with pytest.raises(DependencyNotFoundError, match='.*unknown.*'):
            await f()

        @easy_wrap(arg_dependency=[('x', True, 'a'),
                                   ('y', True, 'b'),
                                   ('z', True, None)])
        async def f(x, y, z):
            pass

        with pytest.raises(TypeError):
            await f()



import pytest


@pytest.mark.compiled_only
def test_fast_dict():
    from antidote.core.container import DependencyCache

    d = DependencyCache()
    n = 37
    for i in range(n):
        d[i] = i ** 2

    for i in range(n):
        assert d[i] == i ** 2

    assert len(d) == n



import pytest

from antidote import Scope, world
from antidote._providers import FactoryProvider
from antidote._providers.service import Parameterized
from antidote.exceptions import (DependencyNotFoundError, DuplicateDependencyError,
                                 FrozenWorldError)


class A:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


def build(**kwargs) -> A:
    return A(**kwargs)


@pytest.fixture()
def provider():
    with world.test.empty():
        world.test.singleton('lazy_build', build)
        world.provider(FactoryProvider)
        yield world.get(FactoryProvider)


@pytest.fixture(params=[dict(factory=build), dict(factory_dependency='lazy_build')])
def factory_params(request):
    return request.param


@pytest.fixture(params=[
    pytest.param(None, id='scope=None'),
    pytest.param(Scope.singleton(), id='scope=singleton')
])
def scope(request):
    return request.param


def test_str(provider: FactoryProvider, scope: Scope):
    provider.register(A, factory=build, scope=scope)
    assert 'A' in str(provider)
    assert 'build' in str(provider)


def test_simple(provider: FactoryProvider):
    factory_id = provider.register(A, factory=build, scope=Scope.singleton())
    assert isinstance(world.get(factory_id), A)
    # singleton
    assert world.get(factory_id) is world.get(factory_id)


def test_lazy(provider: FactoryProvider):
    world.test.singleton('A', build)
    factory_id = provider.register(A, factory_dependency='A', scope=Scope.singleton())
    assert isinstance(world.get(factory_id), A)
    # singleton
    assert world.get(factory_id) is world.get(factory_id)


def test_exists(scope: Scope, factory_params: dict):
    provider = FactoryProvider()
    factory_id = provider.register(A, scope=scope, **factory_params)

    assert not provider.exists(object())
    assert not provider.exists(Parameterized(object(), parameters=dict(a=1)))
    assert provider.exists(factory_id)
    assert provider.exists(Parameterized(factory_id, parameters=dict(a=1)))
    assert not provider.exists(A)
    assert not provider.exists(build)


@pytest.mark.parametrize('singleton', [True, False])
def test_singleton(provider: FactoryProvider, singleton: bool, factory_params: dict):
    scope = Scope.singleton() if singleton else None
    factory_id = provider.register(A, scope=scope, **factory_params)
    assert isinstance(world.get(factory_id), A)
    assert (world.get(factory_id) is world.get(factory_id)) == singleton


def test_multiple_factories(provider: FactoryProvider, scope: Scope):
    def build2(**kwargs) -> A:
        return A(**kwargs)

    b = provider.register(A, factory=build, scope=scope)
    b2 = provider.register(A, factory=build2, scope=scope)

    assert isinstance(world.get(b), A)
    assert isinstance(world.get(b2), A)


def test_duplicate_dependency(provider: FactoryProvider,
                              scope: Scope,
                              factory_params: dict):
    provider.register(A, scope=scope, **factory_params)
    with pytest.raises(DuplicateDependencyError):
        provider.register(A, scope=scope, **factory_params)


def test_parameterized_dependency(scope: Scope, factory_params: dict):
    with world.test.empty():
        world.test.singleton('lazy_build', build)
        provider = FactoryProvider()
        factory_id = provider.register(A, scope=scope, **factory_params)
        kwargs = dict(test=object())
        a = world.test.maybe_provide_from(provider,
                                          Parameterized(factory_id, kwargs)).unwrapped
        assert isinstance(a, A)
        assert a.kwargs == kwargs


def test_copy(scope: Scope):
    class B:
        pass

    class C:
        pass

    def build2() -> A:
        return A(build2=True)

    with world.test.empty():
        original = FactoryProvider()
        a_id = original.register(A, factory=build, scope=scope)

        # cloned has the same dependencies
        cloned = original.clone(False)
        with pytest.raises(DuplicateDependencyError):
            cloned.register(A, factory=build, scope=scope)
        assert isinstance(world.test.maybe_provide_from(cloned, a_id).unwrapped, A)

        # Adding dependencies to either original or cloned, should not impact the
        # other one.
        _ = original.register(B, factory=lambda: B(), scope=scope)
        _ = cloned.register(B, factory=lambda: B(), scope=scope)

        cloned.register(C, factory=lambda: C(), scope=scope)
        original.register(C, factory=lambda: C(), scope=scope)

    with world.test.empty():
        world.test.singleton('build', build)
        original = FactoryProvider()
        a_id = original.register(A, factory_dependency='build', scope=scope)
        a = world.test.maybe_provide_from(original, a_id).unwrapped

        assert isinstance(a, A)
        assert a.kwargs == {}

        cloned_with_singletons = original.clone(True)
        cloned = original.clone(False)
        with world.test.empty():
            world.test.singleton('build', build2)
            a2 = world.test.maybe_provide_from(cloned, a_id).unwrapped

            assert isinstance(a2, A)
            assert a2.kwargs == dict(build2=True)

            # We kept singletons, so previous dependency 'build' has been kept.
            a3 = world.test.maybe_provide_from(cloned_with_singletons, a_id).unwrapped
            assert isinstance(a3, A)
            assert a3.kwargs == {}


def test_freeze(provider: FactoryProvider, scope: Scope):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        provider.register(A, factory=build, scope=scope)


def test_factory_id_repr(provider: FactoryProvider, scope: Scope):
    factory_id = provider.register(A, factory=build, scope=scope)
    assert f"{__name__}.A" in repr(factory_id)
    assert f"{__name__}.build" in repr(factory_id)

    class B:
        pass

    factory_id = provider.register(B, factory_dependency=build, scope=scope)
    assert f"{__name__}.test_factory_id_repr.<locals>.B" in repr(factory_id)
    assert f"{__name__}.build" in repr(factory_id)


def test_unknown_dependency(scope: Scope):
    p1 = FactoryProvider()
    assert world.test.maybe_provide_from(p1, object()) is None
    assert p1.maybe_debug(object()) is None

    p2 = FactoryProvider()
    dependency = p2.register(A, factory=build, scope=scope)
    assert world.test.maybe_provide_from(p1, dependency) is None
    assert p1.maybe_debug(dependency) is None


def test_invalid_lazy_dependency(scope: Scope):
    provider = FactoryProvider()
    fid = provider.register(A, factory_dependency="factory", scope=scope)

    with pytest.raises(DependencyNotFoundError, match=".*factory.*"):
        world.test.maybe_provide_from(provider, fid)


@pytest.mark.parametrize('output, factory, scope', [
    pytest.param(object(), build, None, id='output'),
    pytest.param(A, object(), None, id='factory'),
    pytest.param(A, build, object(), id='scope')
])
def test_sanity_checks(provider: FactoryProvider, output, factory, scope):
    with pytest.raises((AssertionError, TypeError)):
        provider.register(output, factory=factory, scope=scope)


def test_custom_scope(provider: FactoryProvider):
    dummy_scope = world.scopes.new(name='dummy')

    class MyService:
        pass

    fid = provider.register(MyService, factory=lambda: MyService(), scope=dummy_scope)
    my_service = world.get(fid)
    assert my_service is world.get(fid)
    world.scopes.reset(dummy_scope)
    assert my_service is not world.get(fid)



from typing import (Callable, Dict, Hashable, Mapping,
                    Optional,
                    Tuple)

from .._internal import API
from .._internal.utils import debug_repr
from ..core import Container, DependencyDebug, DependencyValue, Provider, Scope


@API.private
class WorldTestProvider(Provider[Hashable]):
    def __init__(self) -> None:
        super().__init__()
        self.__singletons: Dict[Hashable, object] = dict()
        self.__factories: Dict[Hashable,
                               Tuple[Callable[[], object], Optional[Scope]]] = dict()

    def clone(self, keep_singletons_cache: bool) -> 'WorldTestProvider':
        p = WorldTestProvider()
        if keep_singletons_cache:
            p.__singletons = self.__singletons
        p.__factories = self.__factories
        return p

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.__singletons or dependency in self.__factories

    def debug(self, dependency: Hashable) -> DependencyDebug:
        try:
            value = self.__singletons[dependency]
        except KeyError:
            (factory, scope) = self.__factories[dependency]
            return DependencyDebug(f"Singleton: {debug_repr(dependency)} "
                                   f"-> {factory}",
                                   wired=[factory],
                                   scope=scope)
        else:
            return DependencyDebug(f"Singleton: {debug_repr(dependency)} "
                                   f"-> {value}",
                                   scope=Scope.singleton())

    def provide(self, dependency: Hashable, container: Container) -> DependencyValue:
        try:
            value = self.__singletons[dependency]
        except KeyError:
            (factory, scope) = self.__factories[dependency]
            return DependencyValue(factory(), scope=scope)
        else:
            return DependencyValue(value, scope=Scope.singleton())

    def add_singletons(self, dependencies: Mapping[Hashable, object]) -> None:
        for k, v in dependencies.items():
            self._assert_not_duplicate(k)
        self.__singletons.update(dependencies)

    def add_factory(self,
                    dependency: Hashable,
                    *,
                    factory: Callable[[], object],
                    scope: Optional[Scope]) -> None:
        self._assert_not_duplicate(dependency)
        self.__factories[dependency] = (factory, scope)



import pytest

from antidote._internal import state


def test_double_init():
    state.init()
    state.init()
    assert state.current_container() is not None


def test_reset():
    state.reset()
    with pytest.raises(AssertionError):
        state.current_container()

    state.init()
    assert state.current_container() is not None


def test_overridable_container():
    with pytest.raises(RuntimeError):
        state.current_overridable_container()



import collections.abc as c_abc
import inspect
from typing import (Any, Callable, Hashable, Iterable, Mapping, Optional, Sequence,
                    TypeVar, Union, overload)

from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import FinalImmutable

F = TypeVar('F', bound=Callable[..., Any])
AnyF = Union[Callable[..., Any], staticmethod, classmethod]


@API.public
@final
class Arg(FinalImmutable):
    """
    Represents an argument (name and type hint) if you need a very custom injection
    logic.
    """
    __slots__ = ('name', 'type_hint', 'type_hint_with_extras')
    name: str
    """Name of the argument"""
    type_hint: Any
    """Type hint of the argument if any"""
    type_hint_with_extras: Any
    """
    Type hint of the argument if any with include_extras=True, so with annotations.
    """


# API.experimental
DEPENDENCIES_TYPE = Optional[Union[
    Mapping[str, Hashable],  # {arg_name: dependency, ...}
    Sequence[Optional[Hashable]],  # (dependency for arg 1, ...)
    Callable[[Arg], Optional[Hashable]],  # arg -> dependency
]]
AUTO_PROVIDE_TYPE = Optional[Union[
    bool,  # all class type hints or nothing
    Iterable[type],  # specific list of classes
    Callable[[type], bool]  # Function determining which classes should be auto provided
]]


@overload
def inject(__arg: staticmethod,  # noqa: E704  # pragma: no cover
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> staticmethod: ...


@overload
def inject(__arg: classmethod,  # noqa: E704  # pragma: no cover
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> classmethod: ...


@overload
def inject(__arg: F,  # noqa: E704  # pragma: no cover
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> F: ...


@overload
def inject(*,  # noqa: E704  # pragma: no cover
           dependencies: DEPENDENCIES_TYPE = None,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> Callable[[F], F]: ...


@overload
def inject(__arg: Sequence[Hashable],  # noqa: E704  # pragma: no cover
           *,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> Callable[[F], F]: ...


@overload
def inject(__arg: Mapping[str, Hashable],  # noqa: E704  # pragma: no cover
           *,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> Callable[[F], F]: ...


@API.public
def inject(__arg: Union[AnyF, Sequence[Hashable], Mapping[str, Hashable]] = None,
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           auto_provide: AUTO_PROVIDE_TYPE = None,
           strict_validation: bool = True
           ) -> AnyF:
    """
    Inject the dependencies into the function lazily, they are only retrieved
    upon execution. As several options can apply to the same argument, the priority is
    defined as:

    1. Annotated type hints (PEP-593)
    2. Dependencies declared explicitly if any declared with :code:`dependencies`.
    3. Class type hints if specified with :code:`auto_provide`.

    .. doctest:: core_inject

        >>> from antidote import inject, Service, Provide
        >>> class A(Service):
        ...     pass
        >>> class B(Service):
        ...     pass
        >>> # PEP-593 annotation
        ... @inject
        ... def f(a: Provide[A]):
        ...     pass # a = world.get(A)
        >>> # All possibilities for dependency argument.
        >>> @inject(dependencies=dict(b='dependency'))
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject(dict(b='dependency'))  # shortcut
        ... def f(a, b):
        ...     pass
        >>> @inject(dependencies=[None, 'dependency'])  # `a` is ignored
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject([None, 'dependency'])  # shortcut
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject(dependencies=lambda arg: 'dependency' if arg.name == 'b' else None)
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> # All possibilities for auto_provide argument.
        >>> @inject(auto_provide=True)
        ... def f(a: A):
        ...     pass  # a = world.get(A)
        >>> @inject(auto_provide=[B])
        ... def f(a: A, b: B):
        ...     pass  # a, b = <not injected>, world.get(B)
        >>> @inject(auto_provide=False)  # default
        ... def f(a: A):
        ...     pass  # a = <not injected>

    Args:
        __arg: Callable to be wrapped. Can also be used on static methods or class
            methods. May also be sequence of dependencies or mapping from argument
            name to dependencies.
        dependencies: Explicit definition of the dependencies which overrides
            :code:`auto_provide`. Defaults to :py:obj:`None`.
            Can be one of:

            - Mapping from argument name to its dependency
            - Sequence of dependencies which will be mapped with the position
              of the arguments. :py:obj:`None` can be used as a placeholder.
            - Callable which receives :py:class:`~.Arg` as arguments and should
              return the matching dependency. :py:obj:`None` should be used for
              arguments without dependency.
            - String which must have :code:`{arg_name}` as format parameter
        auto_provide: Whether or not class type hints should be used as the arguments
            dependency. Only classes are taken into account and it works with
            :py:class:`~typing.Optional`. An iterable of classes  may also be supplied
            to activate this feature only for those. A function may also be provided, in
            which case it'll be called to determine whether the class type hint ca be
            provided or not. Defaults to :code:`False`.
        strict_validation: Whether arguments should be strictly validated given the
            decorated function's argumnet. For example, a key in the dependencies dict
            that does not match any argument would raise error. Defaults to
            :py:obj:`True`.

    Returns:
        The decorator to be applied or the injected function if the
        argument :code:`func` was supplied.

    """

    from ._injection import raw_inject

    if isinstance(__arg, (c_abc.Sequence, c_abc.Mapping)):
        if isinstance(__arg, str):
            raise TypeError("First argument must be an sequence/mapping of dependencies "
                            "or the function to be wrapped, not a string.")
        if dependencies is not None:
            raise TypeError("dependencies must be None if a sequence/mapping of "
                            "dependencies is provided as first argument.")
        dependencies = __arg
        __arg = None

    def decorate(f: AnyF) -> AnyF:
        return raw_inject(
            f,
            dependencies=dependencies,
            auto_provide=auto_provide if auto_provide is not None else False,
            strict_validation=strict_validation)

    return __arg and decorate(__arg) or decorate


@API.public  # Function will be kept in sync with @inject, so you may use it.
def validate_injection(dependencies: DEPENDENCIES_TYPE = None,
                       auto_provide: AUTO_PROVIDE_TYPE = None) -> None:
    """
    Validates that injection parameters are valid. If not, an error is raised.

    .. doctest:: core_injection_validate_injection

        >>> from antidote.utils import validate_injection
        >>> validate_injection(auto_provide=True)
        >>> validate_injection(auto_provide=object())
        Traceback (most recent call last):
          File "<stdin>", line 1, in ?
        TypeError: auto_provide must be either a boolean or an iterable of classes ...


    Args:
        dependencies: As defined by :py:func:`~.injection.inject`
        auto_provide: As defined by  :py:func:`~.injection.inject`
    """
    if not (dependencies is None
            or (isinstance(dependencies, (c_abc.Sequence, c_abc.Mapping))
                and not isinstance(dependencies, str))
            or callable(dependencies)):
        raise TypeError(
            f"dependencies can be None, a mapping of arguments names to dependencies, "
            f"a sequence of dependencies, a function, "
            f"not a {type(dependencies)!r}"
        )
    if isinstance(dependencies, c_abc.Mapping):
        if not all(isinstance(k, str) for k in dependencies.keys()):
            raise TypeError("Dependencies keys must be argument names (str)")

    if isinstance(auto_provide, str) \
            or not (auto_provide is None
                    or callable(auto_provide)
                    or isinstance(auto_provide, (bool, c_abc.Iterable))):
        raise TypeError(
            f"auto_provide must be either a boolean or an iterable of classes, "
            f"not {type(auto_provide)!r}.")

    # If we can iterate over it safely
    if isinstance(auto_provide, (list, set, tuple, frozenset)):
        for cls in auto_provide:
            if not (isinstance(cls, type) and inspect.isclass(cls)):
                raise TypeError(f"auto_provide must be a boolean or an iterable of "
                                f"classes, but contains {cls!r} which is not a class.")



"""
Similar to the pure Python, but used to have a cdef function for faster access.
"""
import threading
from contextlib import contextmanager
from typing import Callable

# @formatter:off
from antidote.core.container cimport RawContainer
# @formatter:on

cdef:
    RawContainer __container = None
    object __container_lock = threading.RLock()

cdef RawContainer fast_get_container():
    assert __container is not None
    return __container

def current_container() -> RawContainer:
    return fast_get_container()


def current_overridable_container():
    from ..core.container import OverridableRawContainer
    c = fast_get_container()
    if not isinstance(c, OverridableRawContainer):
        raise RuntimeError("Current world does not support overrides. "
                           "Consider using world.test.clone(override=True)")
    return c

def reset():
    global __container
    __container = None

def init():
    global __container
    if __container is None:
        with __container_lock:
            if __container is None:
                from antidote._internal.world import new_container
                __container = new_container()

@contextmanager
def override(create: Callable[[RawContainer], RawContainer]):
    global __container
    with __container_lock:
        old = __container
        try:
            __container = create(old)
            yield
        finally:
            __container = old



from contextlib import contextmanager

import pytest

from antidote import Service, Wiring, service, world, Provide
from antidote._providers import ServiceProvider
from antidote.exceptions import DuplicateDependencyError


@contextmanager
def does_not_raise():
    yield


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        yield


def test_register_simple():
    @service
    class A:
        pass

    assert isinstance(world.get(A), A)
    # singleton by default
    assert world.get(A) is world.get(A)


def test_simple():
    class A(Service):
        pass

    assert isinstance(world.get(A), A)
    # singleton by default
    assert world.get(A) is world.get(A)


def test_register_singleton():
    @service(singleton=True)
    class Singleton:
        pass

    assert world.get(Singleton) is world.get(Singleton)

    @service(singleton=False)
    class NoScope:
        pass

    assert world.get(NoScope) != world.get(NoScope)


def test_singleton():
    class Singleton(Service):
        __antidote__ = Service.Conf(singleton=True)

    assert world.get(Singleton) is world.get(Singleton)

    class NoScope(Service):
        __antidote__ = Service.Conf(singleton=False)

    assert world.get(NoScope) != world.get(NoScope)


def test_custom_scope():
    dummy_scope = world.scopes.new(name='dummy')

    class Scoped(Service):
        __antidote__ = Service.Conf(scope=dummy_scope)

    my_service = world.get(Scoped)
    assert world.get(Scoped) is my_service
    world.scopes.reset(dummy_scope)
    assert world.get(Scoped) is not my_service

    @service(scope=dummy_scope)
    class Scoped:
        pass

    my_service = world.get(Scoped)
    assert world.get(Scoped) is my_service
    world.scopes.reset(dummy_scope)
    assert world.get(Scoped) is not my_service


def test_parameterized():
    x = object()

    class A(Service):
        __antidote__ = Service.Conf(parameters=['x'])

        def __init__(self, **kwargs):
            self.kwargs = kwargs

    a = world.get(A.parameterized(x=x))
    assert a.kwargs == dict(x=x)

    with pytest.raises(ValueError, match=".*parameters.*'x'.*"):
        A.parameterized()

    with pytest.raises(ValueError, match=".*parameters.*'x'.*"):
        A.parameterized(unknown='something')


def test_not_parametrized():
    class A(Service):
        def __init__(self, **kwargs):
            self.kwargs = kwargs

    with pytest.raises(RuntimeError, match=".*parameters.*"):
        A.parameterized(x=1)


def test_invalid_with_default_parameters():
    with pytest.raises(ValueError, match=".*default.*"):
        class A(Service):
            __antidote__ = Service.Conf(parameters=['x'])

            def __init__(self, x: str = 'default'):
                self.x = x


def test_invalid_with_injected_parameters():
    class A(Service):
        pass

    with pytest.raises(ValueError, match=".*injected.*class.*A.*"):
        class B(Service):
            __antidote__ = Service.Conf(parameters=['x'])

            def __init__(self, x: Provide[A]):
                self.x = x


def test_duplicate_registration():
    with pytest.raises(DuplicateDependencyError):
        @service
        class Dummy(Service):
            pass


@pytest.mark.parametrize('cls', ['test', object(), lambda: None])
def test_invalid_class(cls):
    with pytest.raises(TypeError):
        service(cls)


@pytest.mark.parametrize(
    'kwargs,expectation',
    [
        (dict(tags=object()), pytest.raises(TypeError, match=".*tags.*")),
        (dict(tags=['test']), pytest.raises(TypeError, match=".*tags.*")),
        (dict(singleton=object()), pytest.raises(TypeError, match=".*singleton.*")),
        (dict(scope=object()), pytest.raises(TypeError, match=".*scope.*")),
        (dict(wiring=object()), pytest.raises(TypeError, match=".*wiring.*")),
    ]
)
def test_invalid_service_args(kwargs, expectation):
    with expectation:
        @service(**kwargs)
        class Dummy:
            method = None


def test_no_subclass_of_service():
    class A(Service):
        pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class B(A):
            pass


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Dummy(Service):
            __antidote__ = object()


@pytest.mark.parametrize('expectation, parameters', [
    (pytest.raises(TypeError), "string"),
    (pytest.raises(TypeError), object()),
    (pytest.raises(TypeError), [1]),
    (does_not_raise(), ['x']),
    (does_not_raise(), []),
    (does_not_raise(), None),
])
def test_conf_parameters(expectation, parameters):
    with expectation:
        class A(Service):
            __antidote__ = Service.Conf(parameters=parameters)


@pytest.mark.parametrize('expectation, kwargs', [
    pytest.param(pytest.raises(TypeError, match=f'.*{arg}.*'),
                 {arg: object()},
                 id=arg)
    for arg in ['wiring',
                'singleton',
                'scope',
                'parameters']
])
def test_invalid_conf_args(kwargs, expectation):
    with expectation:
        Service.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(singleton=False),
    dict(scope=None),
    dict(wiring=Wiring(methods=['method'])),
    dict(parameters=frozenset(['x']))
])
def test_conf_copy(kwargs):
    conf = Service.Conf(singleton=True).copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v


def test_invalid_copy():
    conf = Service.Conf()
    with pytest.raises(TypeError, match=".*both.*"):
        conf.copy(singleton=False, scope=None)


def test_conf_repr():
    conf = Service.Conf()
    assert "scope" in repr(conf)



from antidote._internal.utils.slots import SlotsRepr


class DummySlot(SlotsRepr):
    __slots__ = ('test', '__value')

    def __init__(self, test, value):
        self.test = test
        self.__value = value


def test_slot_repr_mixin():
    assert repr(DummySlot(1, 'test')) == "DummySlot(test=1, __value='test')"



import builtins
import inspect
from typing import Union, Optional

from .annotations import (AntidoteAnnotation, From, FromArg, Get, INJECT_SENTINEL)
from .injection import Arg
from .._compatibility.typing import Annotated, get_args, get_origin
from .._internal import API
from .._internal.argspec import Argument


@API.private
def extract_annotated_dependency(type_hint: object) -> object:
    origin = get_origin(type_hint)

    # Dependency explicitly given through Annotated (PEP-593)
    if origin is Annotated:
        args = get_args(type_hint)
        antidote_annotations = [a
                                for a in getattr(type_hint, "__metadata__", tuple())
                                if isinstance(a, AntidoteAnnotation)]
        if len(antidote_annotations) > 1:
            raise TypeError(f"Multiple AntidoteAnnotation are not supported. "
                            f"Found {antidote_annotations}")
        elif antidote_annotations:
            annotation: AntidoteAnnotation = antidote_annotations[0]
            if annotation is INJECT_SENTINEL:
                return args[0]
            elif isinstance(annotation, Get):
                return annotation.dependency
            elif isinstance(annotation, From):
                return args[0] @ annotation.source
            else:
                raise TypeError(f"Annotation {annotation} cannot be used"
                                f"outside of a function.")
        else:
            return args[0]

    return type_hint


_BUILTINS_TYPES = {e for e in builtins.__dict__.values() if isinstance(e, type)}


@API.private
def extract_annotated_arg_dependency(argument: Argument) -> object:
    type_hint = argument.type_hint_with_extras
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    # Optional
    if origin is Union:
        if len(args) == 2:
            if isinstance(None, args[1]) or isinstance(None, args[0]):
                type_hint = args[0] if isinstance(None, args[1]) else args[0]
                origin = get_origin(type_hint)
                args = get_args(type_hint)

    # Dependency explicitly given through Annotated (PEP-593)
    if origin is Annotated:
        antidote_annotations = [a
                                for a in type_hint.__metadata__
                                if isinstance(a, AntidoteAnnotation)]
        if len(antidote_annotations) > 1:
            raise TypeError(f"Multiple AntidoteAnnotation are not supported. "
                            f"Found {antidote_annotations}")
        elif antidote_annotations:
            # If antidote annotation, no additional check is done we just return
            # what was specified.
            annotation: AntidoteAnnotation = antidote_annotations[0]
            if annotation is INJECT_SENTINEL:
                return args[0]
            elif isinstance(annotation, Get):
                return annotation.dependency
            elif isinstance(annotation, From):
                return args[0] @ annotation.source
            elif isinstance(annotation, FromArg):
                arg = Arg(argument.name,
                          argument.type_hint,
                          argument.type_hint_with_extras)
                return annotation.function(arg)  # type: ignore
            else:
                raise TypeError(f"Unsupported AntidoteAnnotation, {type(annotation)}")

    return None


@API.private
def extract_auto_provided_arg_dependency(argument: Argument) -> Optional[type]:
    type_hint = argument.type_hint_with_extras
    origin = get_origin(type_hint)
    args = get_args(type_hint)

    # Optional
    if origin is Union:
        if len(args) == 2:
            if isinstance(None, args[1]) or isinstance(None, args[0]):
                type_hint = args[0] if isinstance(None, args[1]) else args[0]
                origin = get_origin(type_hint)
                args = get_args(type_hint)

    dependency = type_hint

    if origin is Annotated:
        antidote_annotations = [a
                                for a in type_hint.__metadata__
                                if isinstance(a, AntidoteAnnotation)]
        if not antidote_annotations:
            dependency = args[0]
        else:
            return None  # necessary for Python 3.6

    if (getattr(dependency, '__module__', '') != 'typing'
            and dependency not in _BUILTINS_TYPES
            and (isinstance(dependency, type) and inspect.isclass(dependency))):
        return dependency

    return None



import inspect
from collections import abc as c_abc
from typing import (FrozenSet, Iterable, Optional, Callable, Any)

from ._internal import API
from ._internal.wrapper import get_wrapper_injections


@API.private
def validated_parameters(parameters: Optional[Iterable[str]]) -> Optional[FrozenSet[str]]:
    if parameters is None:
        return None

    if not isinstance(parameters, str) and isinstance(parameters, c_abc.Iterable):
        parameters = frozenset(parameters)
        if not parameters:
            return None

        if all(isinstance(p, str) for p in parameters):
            return parameters

    raise TypeError(f"parameters must be an iterable of strings or None, "
                    f"not {type(parameters)}")


@API.private
def validate_method_parameters(method: Callable[..., Any],
                               parameters: Optional[FrozenSet[str]]) -> None:
    if parameters is None:
        return

    try:
        injections = get_wrapper_injections(method)
    except TypeError:
        injections = {}

    signature = inspect.signature(method)
    for param in parameters:
        argument = signature.parameters.get(param)
        if argument is not None and argument.default is not inspect.Parameter.empty:
            raise ValueError(f"Parameter '{param}' cannot have a "
                             f"default value in {method.__name__}().")
        if param in injections:
            raise ValueError(f"Parameter '{param}' cannot have an injection in "
                             f"{method.__name__}().  It currently will be injected with "
                             f"{injections[param]!r}")



import pytest

from antidote._internal.utils.immutable import FinalImmutableMeta
from antidote._internal.utils.meta import AbstractMeta, FinalMeta


def test_abstract_meta():
    class A(metaclass=AbstractMeta):
        pass

    with pytest.raises(TypeError, match="abstract"):
        class B(A):
            pass

    class C(metaclass=AbstractMeta, abstract=True):
        pass

    class D(C):
        pass

    with pytest.raises(TypeError, match="abstract"):
        class E(D):
            pass


@pytest.mark.parametrize('meta', [FinalMeta, FinalImmutableMeta])
def test_final_meta(meta):
    class Mixin:
        pass

    class Dummy(Mixin, metaclass=meta):
        __slots__ = ()

    with pytest.raises(TypeError):
        class SubDummy(Dummy):
            __slots__ = ()



from cpython.object cimport PyObject

ctypedef Py_ssize_t Py_hash_t

cdef class DependencyStack:
    cdef:
        PyObject** _trace
        Py_hash_t* _hashes
        size_t _depth
        size_t _capacity

    cdef Exception reset_with_error(self, PyObject* dependency)
    cdef bint push(self, PyObject* dependency)
    cdef void pop(self)



from typing import TypeVar

T = TypeVar('T')


def public(x: T) -> T:
    """
    Objects marked with this decorator are considered to be in the public API.
    Breaking changes will be avoided and will be taken into account in the
    semantic versioning.
    """
    return x


def experimental(x: T) -> T:
    """
    Similar to public, they're part of the public API. However it's not really stable.
    Hence there's a good chance they may be changed or removed in the next release. If
    you're relying on a experimental feature consider opening an issue with your use case
    to advocate its migration to public !
    """
    return x


def private(x: T) -> T:
    """
    Only for internal use. They're are NOT part of the public API, and as such may
    change without warning in later versions. If you rely on private APIs, please open
    an issue.
    """
    return x
